#!/bin/python

"""
This module provides interfacing functions for SYNED and barc4sr
"""

__author__ = ['Rafael Celestre']
__contact__ = 'rafael.celestre@synchrotron-soleil.fr'
__license__ = 'GPL-3.0'
__copyright__ = 'Synchrotron SOLEIL, Saint Aubin, France'
__created__ = '25/NOV/2024'
__changed__ = '25/NOV/2024'

import json
from typing import Any, Dict, Union

import numpy as np

#***********************************************************************************
# functions
#***********************************************************************************

def write_syned_file(json_file: str, light_source_name: str, ElectronBeamClass: object, 
                     MagneticStructureClass: object) -> None:
    """
    Writes a Python dictionary into a SYNED JSON configuration file.

    Parameters:
        json_file (str): The path to the JSON file where the dictionary will be written.
        light_source_name (str): The name of the light source.
        ElectronBeamClass (type): The class representing electron beam parameters.
        MagneticStructureClass (type): The class representing magnetic structure parameters.
    """

    data = {
        "CLASS_NAME": "LightSource",
        "name": light_source_name,
        "electron_beam": vars(ElectronBeamClass),
        "magnetic_structure": vars(MagneticStructureClass)
    }

    with open(json_file, 'w') as file:
        json.dump(data, file, indent=4)


def read_syned_file(json_file: str) -> Dict[str, Any]:
    """
    Reads a SYNED JSON configuration file and returns its contents as a dictionary.

    Parameters:
        json_file (str): The path to the SYNED JSON configuration file.

    Returns:
        dict: A dictionary containing the contents of the JSON file.
    """
    with open(json_file) as f:
        data = json.load(f)
    return data


def syned_dictionary(json_file: str, magnetic_measurement: Union[str, None], observation_point: float, 
                     hor_slit: float, ver_slit: float, hor_slit_cen: float, ver_slit_cen: float) -> dict:
    """
    Generate beamline parameters based on a SYNED JSON configuration file and additional input parameters.

    Args:
        json_file (str): Path to the SYNED JSON configuration file.
        magnetic_measurement (Union[str, None]): Path to the file containing magnetic measurement data.
            Overrides SYNED undulator data if provided.
        observation_point (float): Distance to the observation point in meters.
        hor_slit (float): Horizontal slit size in meters.
        ver_slit (float): Vertical slit size in meters.
        hor_slit_cen (float): Horizontal slit center position in meters.
        ver_slit_cen (float): Vertical slit center position in meters.

    Returns:
        dict: A dictionary containing beamline parameters, including electron beam characteristics,
              magnetic structure details, and radiation observation settings.
    """

    data = read_syned_file(json_file)

    beamline = {}
    # accelerator
    beamline['ElectronEnergy'] = data["electron_beam"]["energy_in_GeV"]
    beamline['ElectronCurrent'] = data["electron_beam"]["current"]
    beamline['ElectronEnergySpread'] = data["electron_beam"]["energy_spread"]
    # electron beam
    beamline['ElectronBeamSizeH'] = np.sqrt(data["electron_beam"]["moment_xx"])
    beamline['ElectronBeamSizeV'] = np.sqrt(data["electron_beam"]["moment_yy"])
    beamline['ElectronBeamDivergenceH'] = np.sqrt(data["electron_beam"]["moment_xpxp"])
    beamline['ElectronBeamDivergenceV'] = np.sqrt(data["electron_beam"]["moment_ypyp"])
    # magnetic structure
    beamline['magnetic_measurement'] = magnetic_measurement
    # undulator        
    if data["magnetic_structure"]["CLASS_NAME"].startswith("U"):
        beamline['NPeriods'] = data["magnetic_structure"]["number_of_periods"]
        beamline['PeriodID'] = data["magnetic_structure"]["period_length"]

        beamline['Kh'] = data["magnetic_structure"]["K_horizontal"]
        beamline['MagFieldPhaseH'] = data["magnetic_structure"]["B_horizontal_phase"]
        beamline['MagFieldSymmetryH'] = data["magnetic_structure"]["B_horizontal_symmetry"]

        beamline['Kv'] = data["magnetic_structure"]["K_vertical"]
        beamline['MagFieldPhaseV'] = data["magnetic_structure"]["B_vertical_phase"]
        beamline['MagFieldSymmetryV'] = data["magnetic_structure"]["B_vertical_symmetry"]
    # bending magnet        
    if data["magnetic_structure"]["CLASS_NAME"].startswith("B"):
        beamline['Bh'] = data["magnetic_structure"]["B_horizontal"]
        beamline['Bv'] = data["magnetic_structure"]["B_vertical"]
        beamline['R'] = data["magnetic_structure"]["radius"]
        beamline['Leff'] = data["magnetic_structure"]["length"]
        beamline['Ledge'] = data["magnetic_structure"]["length_edge"]
    # radiation observation
    beamline['distance'] = observation_point
    beamline['slitH'] = hor_slit
    beamline['slitV'] = ver_slit
    beamline['slitHcenter'] = hor_slit_cen
    beamline['slitVcenter'] = ver_slit_cen
  
    return beamline


def barc4sr_dictionary(light_source: object, magnetic_measurement: Union[str, None], 
                       observation_point: float, hor_slit: float, ver_slit: float, 
                       hor_slit_cen: float, ver_slit_cen: float) -> dict:
    """
    Generate beamline parameters based on a SYNED JSON configuration file and additional input parameters.

    Args:
        light_source (SynchrotronSource): Instance of SynchrotronSource or any object that inherits from it.
        magnetic_measurement (Union[str, None]): Path to the file containing magnetic measurement data.
            Overrides SYNED undulator data if provided.
        observation_point (float): Distance to the observation point in meters.
        hor_slit (float): Horizontal slit size in meters.
        ver_slit (float): Vertical slit size in meters.
        hor_slit_cen (float): Horizontal slit center position in meters.
        ver_slit_cen (float): Vertical slit center position in meters.

    Returns:
        dict: A dictionary containing beamline parameters, including electron beam characteristics,
              magnetic structure details, and radiation observation settings.
    """

    beamline = {}
    # accelerator
    beamline['ElectronEnergy'] = light_source.ElectronBeam.energy_in_GeV
    beamline['ElectronCurrent'] = light_source.ElectronBeam.current
    beamline['ElectronEnergySpread'] = light_source.ElectronBeam.energy_spread
    # electron beam
    beamline['ElectronBeamSizeH'] = np.sqrt(light_source.ElectronBeam.moment_xx)
    beamline['ElectronBeamSizeV'] = np.sqrt(light_source.ElectronBeam.moment_yy)
    beamline['ElectronBeamDivergenceH'] = np.sqrt(light_source.ElectronBeam.moment_xpxp)
    beamline['ElectronBeamDivergenceV'] = np.sqrt(light_source.ElectronBeam.moment_ypyp)
    # magnetic structure
    beamline['magnetic_measurement'] = magnetic_measurement
    # undulator        
    if light_source.MagneticStructure.CLASS_NAME.startswith("U"):
        beamline['NPeriods'] = light_source.MagneticStructure.number_of_periods
        beamline['PeriodID'] = light_source.MagneticStructure.period_length

        beamline['Kh'] = light_source.MagneticStructure.K_horizontal
        beamline['MagFieldPhaseH'] = light_source.MagneticStructure.B_horizontal_phase
        beamline['MagFieldSymmetryH'] = light_source.MagneticStructure.B_horizontal_symmetry

        beamline['Kv'] = light_source.MagneticStructure.K_vertical
        beamline['MagFieldPhaseV'] = light_source.MagneticStructure.B_vertical_phase
        beamline['MagFieldSymmetryV'] = light_source.MagneticStructure.B_vertical_symmetry
    # bending magnet        
    if light_source.MagneticStructure.CLASS_NAME.startswith("B"):
        beamline['Bh'] = light_source.MagneticStructure.B_horizontal
        beamline['Bv'] = light_source.MagneticStructure.B_vertical
        beamline['R'] = light_source.MagneticStructure.radius
        beamline['Leff'] = light_source.MagneticStructure.length
        beamline['Ledge'] = light_source.MagneticStructure.length_edge
    # radiation observation
    beamline['distance'] = observation_point
    beamline['slitH'] = hor_slit
    beamline['slitV'] = ver_slit
    beamline['slitHcenter'] = hor_slit_cen
    beamline['slitVcenter'] = ver_slit_cen
  
    return beamline



#!/bin/python

""" 
This module provides SR classes, SRW interfaced functions, r/w SYNED compatible functions,
r/w functions for the electron trajectory and magnetic field as well as other auxiliary 
functions.
"""
__author__ = ['Rafael Celestre']
__contact__ = 'rafael.celestre@synchrotron-soleil.fr'
__license__ = 'GPL-3.0'
__copyright__ = 'Synchrotron SOLEIL, Saint Aubin, France'
__created__ = '15/MAR/2024'
__changed__ = '04/DEC/2024'

import array
import copy
import multiprocessing as mp
import os
from time import time
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from joblib import Parallel, delayed
from scipy.constants import physical_constants

from barc4sr.aux_energy import get_gamma

try:
    import srwpy.srwlib as srwlib
    USE_SRWLIB = True
except:
    import oasys_srw.srwlib as srwlib
    USE_SRWLIB = True
if USE_SRWLIB is False:
     raise AttributeError("SRW is not available")

PLANCK = physical_constants["Planck constant"][0]
LIGHT = physical_constants["speed of light in vacuum"][0]
CHARGE = physical_constants["atomic unit of charge"][0]
MASS = physical_constants["electron mass"][0]
PI = np.pi

#***********************************************************************************
# SRW interface functions (high level)
#***********************************************************************************

def set_light_source(file_name: str,
                     bl: dict,
                     electron_trajectory: bool,
                     id_type: str,
                     **kwargs) -> Tuple[srwlib.SRWLPartBeam, srwlib.SRWLMagFldC, srwlib.SRWLPrtTrj]:
    """
    Set up the light source parameters including electron beam, magnetic structure, and electron trajectory.

    Args:
        file_name (str): The name of the output file.
        bl (dict): Beamline parameters dictionary containing essential information for setup.
        electron_trajectory (bool): Whether to calculate and save electron trajectory.
        id_type (str): Type of magnetic structure, can be undulator (u), wiggler (w), or bending magnet (bm).

    Optional Args (kwargs):
        ebeam_initial_position (float): Electron beam initial average longitudinal position [m]
        magfield_initial_position (float): Longitudinal position of the magnet center [m]
        magnetic_measurement (str): Path to the file containing magnetic measurement data.
        tabulated_undulator_mthd (int): Method to tabulate the undulator field.

    Returns:
        Tuple[srwlib.SRWLPartBeam, srwlib.SRWLMagFldC, srwlib.SRWLPrtTrj]: A tuple containing the electron beam,
        magnetic structure, and electron trajectory.
    """    

    ebeam_initial_position = kwargs.get('ebeam_initial_position', 0)
    magfield_central_position = kwargs.get('magfield_central_position', 0)
    magnetic_measurement = kwargs.get('magnetic_measurement', None)
    tabulated_undulator_mthd = kwargs.get('tabulated_undulator_mthd', 0)

    # ----------------------------------------------------------------------------------
    # definition of the electron beam
    # ----------------------------------------------------------------------------------
    print('> Generating the electron beam ... ', end='')
    eBeam = set_electron_beam(bl,
                              id_type,
                              initial_position=ebeam_initial_position)
    print('completed')
    # ----------------------------------------------------------------------------------
    # definition of magnetic structure
    # ----------------------------------------------------------------------------------
    print('> Generating the magnetic structure ... ', end='')
    magFldCnt = set_magnetic_structure(bl, 
                                       id_type,
                                       magnetic_measurement = magnetic_measurement, 
                                       magfield_central_position = magfield_central_position,
                                       tabulated_undulator_mthd = tabulated_undulator_mthd)
    print('completed')
    # ----------------------------------------------------------------------------------
    # calculate electron trajectory
    # ----------------------------------------------------------------------------------
    print('> Electron trajectory calculation ... ', end='')
    if electron_trajectory:
        # electron_trajectory_file_name = file_name+"_eTraj.dat"
        eTraj = srwlCalcPartTraj(eBeam, magFldCnt)
        # eTraj.save_ascii(electron_trajectory_file_name)
        # print(f">>>{electron_trajectory_file_name}<<< ", end='')
    else:
        eTraj = 0
    print('completed')
    return eBeam, magFldCnt, eTraj


def set_electron_beam(bl: dict, 
                      id_type: str, 
                      **kwargs) -> srwlib.SRWLPartBeam:
    """
    Set up the electron beam parameters.

    Parameters:
        bl (dict): Dictionary containing beamline parameters.
        id_type (str): Type of magnetic structure, can be undulator (u), wiggler (w), or bending magnet (bm).

    Optional Args (kwargs):
        ebeam_initial_position (float): Electron beam initial average longitudinal position [m]

    Returns:
        srwlib.SRWLPartBeam: Electron beam object initialized with specified parameters.

    """
    initial_position = kwargs.get('initial_position', 0)

    eBeam = srwlib.SRWLPartBeam()
    eBeam.Iavg = bl['ElectronCurrent']  # average current [A]
    eBeam.partStatMom1.x = 0.  # initial transverse positions [m]
    eBeam.partStatMom1.y = 0.
    if id_type.startswith('u'):
        eBeam.partStatMom1.z = - bl['PeriodID'] * (bl['NPeriods'] + 4) / 2  # initial longitudinal positions
    else:
        eBeam.partStatMom1.z = initial_position
    eBeam.partStatMom1.xp = 0  # initial relative transverse divergence [rad]
    eBeam.partStatMom1.yp = 0
    eBeam.partStatMom1.gamma = get_gamma(bl['ElectronEnergy'])

    sigX = bl['ElectronBeamSizeH']  # horizontal RMS size of e-beam [m]
    sigXp = bl['ElectronBeamDivergenceH']  # horizontal RMS angular divergence [rad]
    sigY = bl['ElectronBeamSizeV']  # vertical RMS size of e-beam [m]
    sigYp = bl['ElectronBeamDivergenceV']  # vertical RMS angular divergence [rad]
    sigEperE = bl['ElectronEnergySpread']  

    # 2nd order stat. moments:
    eBeam.arStatMom2[0] = sigX * sigX  # <(x-<x>)^2>
    eBeam.arStatMom2[1] = 0  # <(x-<x>)(x'-<x'>)>
    eBeam.arStatMom2[2] = sigXp * sigXp  # <(x'-<x'>)^2>
    eBeam.arStatMom2[3] = sigY * sigY  # <(y-<y>)^2>
    eBeam.arStatMom2[4] = 0  # <(y-<y>)(y'-<y'>)>
    eBeam.arStatMom2[5] = sigYp * sigYp  # <(y'-<y'>)^2>
    eBeam.arStatMom2[10] = sigEperE * sigEperE  # <(E-<E>)^2>/<E>^2

    return eBeam


def set_magnetic_structure(bl: dict, 
                           id_type: str, 
                           **kwargs) -> srwlib.SRWLMagFldC:
    """
    Sets up the magnetic field container.

    Args:
        bl (dict): Dictionary containing beamline parameters.
        id_type (str): Type of magnetic structure, can be undulator (u), wiggler (w), or bending magnet (bm).

    Optional Args (kwargs):
        magfield_central_position (float): Longitudinal position of the magnet center [m]
        magnetic_measurement (str): Path to the tabulated magnetic field data.
        tabulated_undulator_mthd (int): Method to use for generating undulator field if magnetic_measurement is provided. Defaults to 0

    Returns:
        srwlib.SRWLMagFldC: Magnetic field container.

    """
    magnetic_measurement = kwargs.get('magnetic_measurement', None)
    magfield_central_position = kwargs.get('magfield_central_position', 0)

    if id_type.startswith('u'):
        tabulated_undulator_mthd = kwargs.get('tabulated_undulator_mthd', 0)
        if magnetic_measurement is None:    # ideal sinusoidal undulator magnetic structure
            und = srwlib.SRWLMagFldU()
            und.set_sin(_per=bl["PeriodID"],
                        _len=bl['PeriodID']*bl['NPeriods'], 
                        _bx=bl['Kh']*2*PI*MASS*LIGHT/(CHARGE*bl["PeriodID"]), 
                        _by=bl['Kv']*2*PI*MASS*LIGHT/(CHARGE*bl["PeriodID"]), 
                        _phx=bl['MagFieldPhaseH'], 
                        _phy=bl['MagFieldPhaseV'], 
                        _sx=bl['MagFieldSymmetryH'], 
                        _sy=bl['MagFieldSymmetryV'])

            magFldCnt = srwlib.SRWLMagFldC(_arMagFld=[und],
                                            _arXc=srwlib.array('d', [0.0]),
                                            _arYc=srwlib.array('d', [0.0]),
                                            _arZc=srwlib.array('d', [magfield_central_position]))
            
        else:    # tabulated magnetic field
            magFldCnt = srwlib.srwl_uti_read_mag_fld_3d(magnetic_measurement, _scom='#')
            print(" tabulated magnetic field ... ", end="")
            if tabulated_undulator_mthd  != 0:   # similar to srwl_bl.set_und_per_from_tab()
                # TODO: parametrise
                """Setup periodic Magnetic Field from Tabulated one
                :param _rel_ac_thr: relative accuracy threshold
                :param _max_nh: max. number of harmonics to create
                :param _max_per: max. period length to consider
                """
                _rel_ac_thr=0.05
                _max_nh=50
                _max_per=0.1
                arHarm = []
                for i in range(_max_nh): 
                    arHarm.append(srwlib.SRWLMagFldH())
                magFldCntHarm = srwlib.SRWLMagFldC(srwlib.SRWLMagFldU(arHarm))
                srwlib.srwl.UtiUndFromMagFldTab(magFldCntHarm, magFldCnt, [_rel_ac_thr, _max_nh, _max_per])
                return magFldCntHarm
            
    if id_type.startswith('bm'):

        bm = srwlib.SRWLMagFldM()
        bm.G = bl["Bv"]
        bm.m = 1         # multipole order: 1 for dipole, 2 for quadrupole, 3 for sextupole, 4 for octupole
        bm.n_or_s = 'n'  # normal ('n') or skew ('s')
        bm.Leff = bl["Leff"]
        bm.Ledge = bl["Ledge"]
        bm.R = bl["R"]

        magFldCnt = srwlib.SRWLMagFldC(_arMagFld=[bm],
                                       _arXc=srwlib.array('d', [0.0]),
                                       _arYc=srwlib.array('d', [0.0]),
                                       _arZc=srwlib.array('d', [magfield_central_position]))

    return magFldCnt

#***********************************************************************************
# SRW interface functions (low level level)
#***********************************************************************************

def srwlCalcPartTraj(eBeam:srwlib.SRWLPartBeam,
                     magFldCnt: srwlib.SRWLMagFldC,
                     number_points: int = 50000, 
                     ctst: float = 0, 
                     ctfi: float = 0) -> srwlib.SRWLPrtTrj:
    """
    Calculate the trajectory of an electron through a magnetic field.

    Args:
        eBeam (srwlib.SRWLPartBeam): Particle beam properties.
        magFldCnt (srwlib.SRWLMagFldC): Magnetic field container representing the magnetic field.
        number_points (int, optional): Number of points for trajectory calculation. Defaults to 50000.
        ctst (float, optional): Initial time (ct) for trajectory calculation. Defaults to 0.
        ctfi (float, optional): Final time (ct) for trajectory calculation. Defaults to 0.

    Returns:
        srwlib.SRWLPrtTrj: Object containing the calculated trajectory.
    """
    partTraj = srwlib.SRWLPrtTrj()
    partTraj.partInitCond = eBeam.partStatMom1
    partTraj.allocate(number_points, True)
    partTraj.ctStart = ctst
    partTraj.ctEnd = ctfi

    arPrecPar = [1] 
    srwlib.srwl.CalcPartTraj(partTraj, magFldCnt, arPrecPar)

    return partTraj


def srwlibCalcElecFieldSR(bl: dict, 
                          eBeam: srwlib.SRWLPartBeam, 
                          magFldCnt: srwlib.SRWLMagFldC, 
                          energy_array: np.ndarray,
                          h_slit_points: int, 
                          v_slit_points: int, 
                          radiation_characteristic: int, 
                          radiation_dependence: int, 
                          radiation_polarisation: int,
                          id_type: str,
                          parallel: bool,
                          num_cores: int=None) -> np.ndarray:
    """
    Calculates the electric field for synchrotron radiation.

    Args:
        bl (dict): Dictionary containing beamline parameters.
        eBeam (srwlib.SRWLPartBeam): Electron beam properties.
        magFldCnt (srwlib.SRWLMagFldC): Magnetic field container.
        energy_array (np.ndarray): Array of photon energies [eV].
        h_slit_points (int): Number of horizontal slit points.
        v_slit_points (int): Number of vertical slit points.
        radiation_characteristic (int): Radiation characteristic:
               =0 -"Single-Electron" Intensity; 
               =1 -"Multi-Electron" Intensity; 
               =4 -"Single-Electron" Radiation Phase; 
               =5 -Re(E): Real part of Single-Electron Electric Field;
               =6 -Im(E): Imaginary part of Single-Electron Electric Field
        radiation_dependence (int): Radiation dependence (e.g., 1 for angular distribution).
               =0 -vs e (photon energy or time);
               =1 -vs x (horizontal position or angle);
               =2 -vs y (vertical position or angle);
               =3 -vs x&y (horizontal and vertical positions or angles);
               =4 -vs e&x (photon energy or time and horizontal position or angle);
               =5 -vs e&y (photon energy or time and vertical position or angle);
               =6 -vs e&x&y (photon energy or time, horizontal and vertical positions or angles);
        radiation_polarisation (int): Polarisation component to be extracted.
               =0 -Linear Horizontal; 
               =1 -Linear Vertical; 
               =2 -Linear 45 degrees; 
               =3 -Linear 135 degrees; 
               =4 -Circular Right; 
               =5 -Circular Left; 
               =6 -Total
        id_type (str): Type of magnetic structure, can be undulator (u), wiggler (w), or bending magnet (bm).
        parallel (bool): Whether to use parallel computation.
        num_cores (int, optional): Number of CPU cores to use for parallel computation. If not specified, 
                            it defaults to the number of available CPU cores.

    Returns:
        np.ndarray: Array containing intensity data, horizontal and vertical axes
    """
    
    arPrecPar = [0]*7
    if id_type.startswith('bm') or id_type.startswith('w'):
        arPrecPar[0] = 2      # SR calculation method: 0- "manual", 1- "auto-undulator", 2- "auto-wiggler"
    else:
        arPrecPar[0] = 1
    arPrecPar[1] = 0.001  
    arPrecPar[2] = 0     # longitudinal position to start integration (effective if < zEndInteg)
    arPrecPar[3] = 0     # longitudinal position to finish integration (effective if > zStartInteg)
    arPrecPar[4] = 50000 # Number of points for trajectory calculation
    arPrecPar[5] = 1     # Use "terminating terms"  or not (1 or 0 respectively)
    arPrecPar[6] = 0     # sampling factor for adjusting nx, ny (effective if > 0)

    if num_cores is None:
        num_cores = mp.cpu_count()

    if parallel:
        dE = np.diff(energy_array)    
        dE1 = np.min(dE)
        dE2 = np.max(dE)

        wiggler_regime = bool(energy_array[-1]>51*energy_array[0])

        # if np.allclose(dE1, dE2) and wiggler_regime:
        if wiggler_regime:
            chunk_size = 20
            n_slices = len(energy_array)

            chunks = [(energy_array[i:i + chunk_size],
                    bl, 
                    eBeam,
                    magFldCnt, 
                    arPrecPar, 
                    h_slit_points, 
                    v_slit_points, 
                    radiation_characteristic, 
                    radiation_dependence,
                    radiation_polarisation,
                    parallel) for i in range(0, n_slices, chunk_size)]
            
            with mp.Pool() as pool:
                results = pool.map(core_srwlibCalcElecFieldSR, chunks)
        else:
            dE = (energy_array[-1] - energy_array[0]) / num_cores
            energy_chunks = []

            for i in range(num_cores):
                bffr = copy.copy(energy_array)                
                bffr = np.delete(bffr, bffr < dE * (i) + energy_array[0])
                if i + 1 != num_cores:
                    bffr = np.delete(bffr, bffr >= dE * (i + 1) + energy_array[0])
                energy_chunks.append(bffr)

            results = Parallel(n_jobs=num_cores)(delayed(core_srwlibCalcElecFieldSR)((
                                                                        list_pairs,
                                                                        bl,
                                                                        eBeam,
                                                                        magFldCnt,
                                                                        arPrecPar,
                                                                        h_slit_points,
                                                                        v_slit_points,
                                                                        radiation_characteristic,
                                                                        radiation_dependence,
                                                                        radiation_polarisation,
                                                                        parallel))
                                                for list_pairs in energy_chunks)
            
        for i, (intensity_chunck, h_chunck, v_chunck, e_chunck, t_chunck) in enumerate(results):
            if i == 0:
                intensity = intensity_chunck
                energy_array = np.asarray([e_chunck[0]])
                energy_chunks = np.asarray([len(e_chunck)])
                time_array = np.asarray([t_chunck])
            else:
                intensity = np.concatenate((intensity, intensity_chunck), axis=0)
                energy_array = np.concatenate((energy_array, np.asarray([e_chunck[0]])))
                energy_chunks = np.concatenate((energy_chunks, np.asarray([len(e_chunck)])))
                time_array = np.concatenate((time_array, np.asarray([t_chunck])))

        if not wiggler_regime:
            print(">>> ellapse time:")
            for ptime in range(len(time_array)):
                print(f" Core {ptime+1}: {time_array[ptime]:.2f} s for {energy_chunks[ptime]} pts (E0 = {energy_array[ptime]:.1f} eV).")

    else:
        results = core_srwlibCalcElecFieldSR((energy_array,
                                             bl, 
                                             eBeam,
                                             magFldCnt, 
                                             arPrecPar, 
                                             h_slit_points, 
                                             v_slit_points, 
                                             radiation_characteristic, 
                                             radiation_dependence,
                                             radiation_polarisation,
                                             parallel))
        intensity = results[0]

    if h_slit_points == 1 or v_slit_points == 1:
        x_axis = np.asarray([0])
        y_axis = np.asarray([0])
    else:
        x_axis = np.linspace(-bl['slitH']/2-bl['slitHcenter'], bl['slitH']/2-bl['slitHcenter'], h_slit_points)
        y_axis = np.linspace(-bl['slitV']/2-bl['slitVcenter'], bl['slitV']/2-bl['slitVcenter'], v_slit_points)

    return intensity, x_axis, y_axis


def core_srwlibCalcElecFieldSR(args: Tuple[np.ndarray, 
                                           dict, 
                                           srwlib.SRWLPartBeam, 
                                           srwlib.SRWLMagFldC, 
                                           List[float], 
                                           int, int, int, int, int, bool]) -> Tuple[np.ndarray, np.ndarray, np.ndarray, float]:
    """
    Core function to calculate electric field for synchrotron radiation.

    Args:
        args (Tuple): Tuple containing the following elements:
            energy_array (np.ndarray): Array of photon energies [eV].
            bl (dict): Dictionary containing beamline parameters.
            eBeam (srwlib.SRWLPartBeam): Electron beam properties.
            magFldCnt (srwlib.SRWLMagFldC): Magnetic field container.
            arPrecPar (List[float]): Array of parameters for SR calculation.
            h_slit_points (int): Number of horizontal slit points.
            v_slit_points (int): Number of vertical slit points.
            rad_characteristic (int): Radiation characteristic (e.g., 0 for intensity).
            rad_dependence (int): Radiation dependence (e.g., 1 for angular distribution).
            radiation_polarisation (int): Polarisation component to be extracted.
            parallel (bool): Whether to use parallel computation.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray, float]: Tuple containing intensity data, 
                                                          horizontal axis, vertical axis, 
                                                          and computation time.
    """

    energy_array, bl, eBeam, magFldCnt, arPrecPar,  h_slit_points, v_slit_points, \
        rad_characteristic, rad_dependence, rad_polarisation, parallel = args
    
    tzero = time()

    _inPol = rad_polarisation
    _inIntType = rad_characteristic
    _inDepType = rad_dependence

    monochromatic = False
    if isinstance(energy_array, int) or isinstance(energy_array, float):
        monochromatic = True 

    if h_slit_points == 1 or v_slit_points == 1:
        hAxis = np.asarray([0])
        vAxis = np.asarray([0])
        _inDepType = 0
        intensity = np.zeros((energy_array.size))
    else:
        hAxis = np.linspace(-bl['slitH']/2-bl['slitHcenter'], bl['slitH']/2-bl['slitHcenter'], h_slit_points)
        vAxis = np.linspace(-bl['slitV']/2-bl['slitVcenter'], bl['slitV']/2-bl['slitVcenter'], v_slit_points)
        _inDepType = 3
        if monochromatic:
            intensity =  np.zeros((vAxis.size, hAxis.size))
        else:
            intensity = np.zeros((energy_array.size, vAxis.size, hAxis.size))

    if parallel:    
        # this is rather convinient for step by step calculations and less memory intensive
        for ie in range(energy_array.size):
            try:
                mesh = srwlib.SRWLRadMesh(energy_array[ie], energy_array[ie], 1,
                                         hAxis[0], hAxis[-1], h_slit_points,
                                         vAxis[0], vAxis[-1], v_slit_points, 
                                         bl['distance'])

                wfr = srwlib.SRWLWfr()
                wfr.allocate(mesh.ne, mesh.nx, mesh.ny)
                wfr.mesh = mesh
                wfr.partBeam = eBeam

                srwlib.srwl.CalcElecFieldSR(wfr, 0, magFldCnt, arPrecPar)
                if _inIntType == 4:
                    arI1 = array.array('d', [0]*wfr.mesh.nx*wfr.mesh.ny)
                else:
                    arI1 = array.array('f', [0]*wfr.mesh.nx*wfr.mesh.ny)

                srwlib.srwl.CalcIntFromElecField(arI1, wfr, _inPol, _inIntType, _inDepType, wfr.mesh.eStart, 0, 0)
                if _inDepType == 0:    # 0 -vs e (photon energy or time);
                    intensity[ie] = np.asarray(arI1, dtype="float64")
                else:
                    # data = np.ndarray(buffer=arI1, shape=(wfr.mesh.ny, wfr.mesh.nx),dtype=arI1.typecode)
                    data = np.asarray(arI1, dtype="float64").reshape((wfr.mesh.ny, wfr.mesh.nx)) 
                    intensity[ie, :, :] = data
            except:
                 raise ValueError("Error running SRW.")
    else:
        try:
            if monochromatic:
                ei = ef = energy_array
                nf = 1
            else:
                ei = energy_array[0]
                ef = energy_array[-1]
                nf = len(energy_array)

            mesh = srwlib.SRWLRadMesh(ei, ef, nf,
                                      hAxis[0], hAxis[-1], h_slit_points,
                                      vAxis[0], vAxis[-1], v_slit_points, 
                                      bl['distance'])
            
            wfr = srwlib.SRWLWfr()
            wfr.allocate(mesh.ne, mesh.nx, mesh.ny)
            wfr.mesh = mesh
            wfr.partBeam = eBeam

            # srwl_bl.calc_sr_se sets eTraj=0 despite having measured magnetic field
            srwlib.srwl.CalcElecFieldSR(wfr, 0, magFldCnt, arPrecPar)

            if _inDepType == 0:    # 0 -vs e (photon energy or time);
                arI1 = array.array('f', [0]*wfr.mesh.ne)
                srwlib.srwl.CalcIntFromElecField(arI1, wfr, _inPol, _inIntType, _inDepType, wfr.mesh.eStart, 0, 0)
                intensity = np.asarray(arI1, dtype="float64")
            else:
                if monochromatic:
                    if _inIntType == 4:
                        arI1 = array.array('d', [0]*wfr.mesh.nx*wfr.mesh.ny)
                    else:
                        arI1 = array.array('f', [0]*wfr.mesh.nx*wfr.mesh.ny)
                    srwlib.srwl.CalcIntFromElecField(arI1, wfr, _inPol, _inIntType, _inDepType, ei, 0, 0)
                    intensity = np.asarray(arI1, dtype="float64").reshape((wfr.mesh.ny, wfr.mesh.nx))
                else:
                    for ie in range(len(energy_array)):
                        if _inIntType == 4:
                            arI1 = array.array('d', [0]*wfr.mesh.nx*wfr.mesh.ny)
                        else:
                            arI1 = array.array('f', [0]*wfr.mesh.nx*wfr.mesh.ny)
                        srwlib.srwl.CalcIntFromElecField(arI1, wfr, _inPol, _inIntType, _inDepType, energy_array[ie], 0, 0)
                        data = np.asarray(arI1, dtype="float64").reshape((wfr.mesh.ny, wfr.mesh.nx)) #np.ndarray(buffer=arI1, shape=(wfr.mesh.ny, wfr.mesh.nx),dtype=arI1.typecode)
                        intensity[ie, :, :] = data

        except:
             raise ValueError("Error running SRW.")

    return intensity, hAxis, vAxis, energy_array, time()-tzero


def srwlibsrwl_wfr_emit_prop_multi_e(bl: dict,
                                     eBeam: srwlib.SRWLPartBeam, 
                                     magFldCnt: srwlib.SRWLMagFldC, 
                                     energy_array: np.ndarray,
                                     h_slit_points: int, 
                                     v_slit_points: int, 
                                     radiation_polarisation: int,
                                     id_type: str,
                                     number_macro_electrons: int, 
                                     aux_file_name: str,
                                     parallel: bool,
                                     num_cores: int=None,
                                     srApprox: int=0):
    """
    Interface function to compute multi-electron emission and propagation through a beamline using SRW.

    Args:
        bl (dict): Dictionary containing beamline parameters.
        eBeam (srwlib.SRWLPartBeam): Electron beam properties.
        magFldCnt (srwlib.SRWLMagFldC): Magnetic field container.
        energy_array (np.ndarray): Array of photon energies [eV].
        h_slit_points (int): Number of horizontal slit points.
        v_slit_points (int): Number of vertical slit points.
        radiation_polarisation (int): Polarisation component to be extracted.
               =0 -Linear Horizontal; 
               =1 -Linear Vertical; 
               =2 -Linear 45 degrees; 
               =3 -Linear 135 degrees; 
               =4 -Circular Right; 
               =5 -Circular Left; 
               =6 -Total
        id_type (str): Type of magnetic structure, can be undulator (u), wiggler (w), or bending magnet (bm).
        number_macro_electrons (int): Total number of macro-electrons.
        aux_file_name (str): Auxiliary file name for saving intermediate data.
        parallel (bool): Whether to use parallel computation.
        num_cores (int, optional): Number of CPU cores to use for parallel computation. If not specified, 
                                   it defaults to the number of available CPU cores.
        srApprox (int): Approximation to be used at multi-electron integration: 
                0- none (i.e. do standard M-C integration over 5D phase space volume of e-beam), 
                1- integrate numerically only over e-beam energy spread and use convolution to treat transverse emittance
    Returns:
        np.ndarray: Array containing intensity data.
    """
    nMacroElecAvgPerProc = 10   # number of macro-electrons / wavefront to average on worker processes
    nMacroElecSavePer = 100     # intermediate data saving periodicity (in macro-electrons)
    if id_type.startswith('bm') or id_type.startswith('w'):
        srCalcMeth = 2          # SR calculation method: 0- "manual", 1- "auto-undulator", 2- "auto-wiggler"
    else:
        srCalcMeth = 1

    srApprox = 0
    srCalcPrec = 0.01           # SR calculation rel. accuracy

    if h_slit_points == 1 or v_slit_points == 1:
        hAxis = np.asarray([0])
        vAxis = np.asarray([0])
    else:
        hAxis = np.linspace(-bl['slitH']/2-bl['slitHcenter'], bl['slitH']/2-bl['slitHcenter'], h_slit_points)
        vAxis = np.linspace(-bl['slitV']/2-bl['slitVcenter'], bl['slitV']/2-bl['slitVcenter'], v_slit_points)

    if num_cores is None:
        num_cores = mp.cpu_count()

    if parallel:
        dE = np.diff(energy_array)    
        dE1 = np.min(dE)
        dE2 = np.max(dE)

        wiggler_regime = bool(energy_array[-1]>51*energy_array[0])

        # if np.allclose(dE1, dE2) and wiggler_regime:
        if wiggler_regime:
            chunk_size = 20
            n_slices = len(energy_array)

            chunks = [(energy_array[i:i + chunk_size],
                        bl,
                        eBeam, 
                        magFldCnt, 
                        h_slit_points, 
                        v_slit_points, 
                        number_macro_electrons, 
                        aux_file_name+'_'+str(i),
                        srCalcMeth,
                        srCalcPrec,
                        srApprox,
                        radiation_polarisation,
                        nMacroElecAvgPerProc,
                        nMacroElecSavePer) for i in range(0, n_slices, chunk_size)]
            
            with mp.Pool() as pool:
                results = pool.map(core_srwlibsrwl_wfr_emit_prop_multi_e, chunks)
        else:
            dE = (energy_array[-1] - energy_array[0]) / num_cores
            energy_chunks = []

            for i in range(num_cores):
                bffr = copy.copy(energy_array)                
                bffr = np.delete(bffr, bffr < dE * (i) + energy_array[0])
                if i + 1 != num_cores:
                    bffr = np.delete(bffr, bffr >= dE * (i + 1) + energy_array[0])
                energy_chunks.append(bffr)

            results = Parallel(n_jobs=num_cores)(delayed(core_srwlibsrwl_wfr_emit_prop_multi_e)((
                                                                        list_pairs,
                                                                        bl,
                                                                        eBeam, 
                                                                        magFldCnt, 
                                                                        h_slit_points, 
                                                                        v_slit_points, 
                                                                        number_macro_electrons, 
                                                                        aux_file_name+'_'+str(list_pairs[0]),
                                                                        srCalcMeth,
                                                                        srCalcPrec,
                                                                        srApprox,
                                                                        radiation_polarisation,
                                                                        nMacroElecAvgPerProc,
                                                                        nMacroElecSavePer))
                                                for list_pairs in energy_chunks)

        for i, (intensity_chunck, e_chunck, t_chunck) in enumerate(results):
            if i == 0:
                intensity = intensity_chunck
                energy_chunck = np.asarray([e_chunck[0]])
                energy_chunks = np.asarray([len(e_chunck)])
                time_array = np.asarray([t_chunck])
            else:
                intensity = np.concatenate((intensity, intensity_chunck), axis=0)
                energy_chunck = np.concatenate((energy_chunck, np.asarray([e_chunck[0]])))
                energy_chunks = np.concatenate((energy_chunks, np.asarray([len(e_chunck)])))
                time_array = np.concatenate((time_array, np.asarray([t_chunck])))

        if not wiggler_regime:
            print(">>> ellapse time:")
            for ptime in range(len(time_array)):
                print(f" Core {ptime+1}: {time_array[ptime]:.2f} s for {energy_chunks[ptime]} pts (E0 = {energy_chunck[ptime]:.1f} eV).")
    else:
        results = core_srwlibsrwl_wfr_emit_prop_multi_e((energy_array,
                                                        bl,
                                                        eBeam, 
                                                        magFldCnt, 
                                                        h_slit_points, 
                                                        v_slit_points, 
                                                        number_macro_electrons, 
                                                        aux_file_name,
                                                        srCalcMeth,
                                                        srCalcPrec,
                                                        srApprox,
                                                        radiation_polarisation,
                                                        nMacroElecAvgPerProc,
                                                        nMacroElecSavePer))
        intensity = np.asarray(results[0], dtype="float64")

    return intensity, hAxis, vAxis


def core_srwlibsrwl_wfr_emit_prop_multi_e(args: Tuple[np.ndarray,
                                                      dict, 
                                                      srwlib.SRWLPartBeam, 
                                                      srwlib.SRWLMagFldC, 
                                                      int, int, int, str, int, float,
                                                      int, int, int, int]) -> Tuple[np.ndarray, float]:
    """
    Core function for computing multi-electron emission and propagation through a beamline using SRW.

    Args:
        args (tuple): Tuple containing arguments:
            - energy_array (np.ndarray): Array of photon energies [eV].
            - bl (dict): Dictionary containing beamline parameters.
            - eBeam (srwlib.SRWLPartBeam): Electron beam properties.
            - magFldCnt (srwlib.SRWLMagFldC): Magnetic field container.
            - h_slit_points (int): Number of horizontal slit points.
            - v_slit_points (int): Number of vertical slit points.
            - number_macro_electrons (int): Total number of macro-electrons.
            - aux_file_name (str): Auxiliary file name for saving intermediate data.
            - srCalcMeth (int): SR calculation method.
            - srCalcPrec (float): SR calculation relative accuracy.
            - srApprox (int): Approximation to be used at multi-electron integration: 
                    0- none (i.e. do standard M-C integration over 5D phase space volume of e-beam), 
                    1- integrate numerically only over e-beam energy spread and use convolution to treat transverse emittance
            - radiation_polarisation (int): Polarisation component to be extracted.
            - nMacroElecAvgPerProc (int): Number of macro-electrons / wavefront to average on worker processes.
            - nMacroElecSavePer (int): Intermediate data saving periodicity (in macro-electrons).

    Returns:
        tuple: A tuple containing intensity data array and the elapsed time.
    """

    energy_array, bl, eBeam, magFldCnt, h_slit_points, v_slit_points, \
        number_macro_electrons, aux_file_name, srCalcMeth, srCalcPrec, srApprox, radiation_polarisation,\
        nMacroElecAvgPerProc, nMacroElecSavePer = args
    
    tzero = time()

    try:    
        
        if isinstance(energy_array, int) or isinstance(energy_array, float):
            monochromatic = True 
            ei = ef = energy_array
            nf = 1
        else:
            monochromatic = False
            ei = energy_array[0]
            ef = energy_array[-1]
            nf = len(energy_array)

        mesh = srwlib.SRWLRadMesh(ei, 
                                  ef, 
                                  nf,
                                  bl['slitHcenter'] - bl['slitH']/2,
                                  bl['slitHcenter'] + bl['slitH']/2, 
                                  h_slit_points,
                                  bl['slitVcenter'] - bl['slitV']/2, 
                                  bl['slitVcenter'] - bl['slitV']/2, 
                                  v_slit_points,
                                  bl['distance'])

        MacroElecFileName = aux_file_name + '_'+ str(int(number_macro_electrons / 1000)).replace('.', 'p') +'k_ME_intensity.dat'

        stk = srwlib.srwl_wfr_emit_prop_multi_e(_e_beam = eBeam, 
                                                _mag = magFldCnt,
                                                _mesh = mesh,
                                                _sr_meth = srCalcMeth,
                                                _sr_rel_prec = srCalcPrec,
                                                _n_part_tot = number_macro_electrons,
                                                _n_part_avg_proc=nMacroElecAvgPerProc, 
                                                _n_save_per=nMacroElecSavePer,
                                                _file_path=MacroElecFileName, 
                                                _sr_samp_fact=-1, 
                                                _opt_bl=None,
                                                _char=0,
                                                _me_approx=srApprox)
    
        os.system('rm %s'% MacroElecFileName)
        me_intensity = np.asarray(stk.to_int(_pol=radiation_polarisation), dtype='float64')

        if h_slit_points != 1 or v_slit_points != 1:
            k = 0
            if monochromatic:
                data = np.zeros((v_slit_points, h_slit_points))
                for iy in range(v_slit_points):
                    for ix in range(h_slit_points):
                        data[iy, ix] = me_intensity[k]
                        k+=1
            else:
                data = np.zeros((len(energy_array), v_slit_points, h_slit_points))
                for iy in range(v_slit_points):
                    for ix in range(h_slit_points):
                        for ie in range(len(energy_array)):
                            data[ie, iy, ix] = me_intensity[k]
                            k+=1
            me_intensity = data

    except:
         raise ValueError("Error running SRW.")

    return (me_intensity, energy_array, time()-tzero)


def srwlibCalcStokesUR(bl: dict, 
                       eBeam: srwlib.SRWLPartBeam, 
                       magFldCnt: srwlib.SRWLMagFldC, 
                       energy_array: np.ndarray, 
                       resonant_energy: float, 
                       h_slit_points: int, 
                       v_slit_points: int, 
                       radiation_polarisation: int,
                       parallel: bool,
                       num_cores: int=None) -> np.ndarray:
    """
    Calculates the Stokes parameters for undulator radiation.

    Args:
        bl (dict): Dictionary containing beamline parameters.
        eBeam (srwlib.SRWLPartBeam): Electron beam properties.
        magFldCnt (srwlib.SRWLMagFldC): Magnetic field container.
        energy_array (np.ndarray): Array of photon energies [eV].
        resonant_energy (float): Resonant energy [eV].
        h_slit_points (int): Number of horizontal slit points.
        v_slit_points (int): Number of vertical slit points.
        radiation_polarisation (int): Polarisation component to be extracted.
               =0 -Linear Horizontal; 
               =1 -Linear Vertical; 
               =2 -Linear 45 degrees; 
               =3 -Linear 135 degrees; 
               =4 -Circular Right; 
               =5 -Circular Left; 
               =6 -Total
        parallel (bool): Whether to use parallel computation.
        num_cores (int, optional): Number of CPU cores to use for parallel computation. If not specified, 
                                   it defaults to the number of available CPU cores.

    Returns:
        np.ndarray: Array containing intensity data.
    """
    
    if parallel:
        if num_cores is None:
            num_cores = mp.cpu_count()

        energy_chunks = np.array_split(list(energy_array), num_cores)

        results = Parallel(n_jobs=num_cores)(delayed(core_srwlibCalcStokesUR)((
                                                                    list_pairs,
                                                                    bl,
                                                                    eBeam,
                                                                    magFldCnt,
                                                                    h_slit_points,
                                                                    v_slit_points,
                                                                    resonant_energy,
                                                                    radiation_polarisation))
                                             for list_pairs in energy_chunks)
        energy_array = []
        time_array = []
        energy_chunks = []

        k = 0
        for calcs in results:
            energy_array.append(calcs[1][0])
            time_array.append(calcs[2])
            energy_chunks.append(len(calcs[0]))
            if k == 0:
                intensity = np.asarray(calcs[0], dtype="float64")
            else:
                intensity = np.concatenate((intensity, np.asarray(calcs[0], dtype="float64")), axis=0)
            k+=1
        print(">>> ellapse time:")

        for ptime in range(len(time_array)):
            print(f" Core {ptime+1}: {time_array[ptime]:.2f} s for {energy_chunks[ptime]} pts (E0 = {energy_array[ptime]:.1f} eV).")

    else:
        results = core_srwlibCalcStokesUR((energy_array,
                                          bl, 
                                          eBeam,
                                          magFldCnt, 
                                          h_slit_points,
                                          v_slit_points,
                                          resonant_energy,
                                          radiation_polarisation))
        
        intensity = np.asarray(results[0], dtype="float64")

    return intensity


def core_srwlibCalcStokesUR(args: Tuple[np.ndarray, 
                                        dict, 
                                        srwlib.SRWLPartBeam, 
                                        srwlib.SRWLMagFldC, 
                                        np.ndarray,
                                        np.ndarray,
                                        float,
                                        int]) -> Tuple[np.ndarray, float]:
    """
    Core function to calculate Stokes parameters for undulator radiation.

    Args:
        args (tuple): Tuple containing arguments:
            - energy_array (np.ndarray): Array of photon energies [eV].
            - bl (dict): Dictionary containing beamline parameters.
            - eBeam (srwlib.SRWLPartBeam): Electron beam properties.
            - magFldCnt (srwlib.SRWLMagFldC): Magnetic field container.
            - h_slit_points (int): Number of horizontal slit points.
            - v_slit_points (int): Number of vertical slit points.
            - resonant_energy (float): Resonant energy [eV].
            - radiation_polarisation (int): Polarisation component to be extracted.

    Returns:
        Tuple[np.ndarray, float]: Tuple containing intensity data and computation time.
    """

    energy_array, bl, eBeam, magFldCnt, h_slit_points, v_slit_points, resonant_energy, radiation_polarisation = args

    tzero = time()

    try:

        arPrecPar = [0]*5   # for spectral flux vs photon energy
        arPrecPar[0] = 1    # initial UR harmonic to take into account
        arPrecPar[1] = get_undulator_max_harmonic_number(resonant_energy, energy_array[-1]) #final UR harmonic to take into account
        arPrecPar[2] = 1.5  # longitudinal integration precision parameter
        arPrecPar[3] = 1.5  # azimuthal integration precision parameter
        arPrecPar[4] = 1    # calculate flux (1) or flux per unit surface (2)

        npts = len(energy_array)
        stk = srwlib.SRWLStokes() 
        stk.allocate(npts, h_slit_points, v_slit_points)     
        stk.mesh.zStart = bl['distance']
        stk.mesh.eStart = energy_array[0]
        stk.mesh.eFin =   energy_array[-1]
        stk.mesh.xStart = bl['slitHcenter'] - bl['slitH']/2
        stk.mesh.xFin =   bl['slitHcenter'] + bl['slitH']/2
        stk.mesh.yStart = bl['slitVcenter'] - bl['slitV']/2
        stk.mesh.yFin =   bl['slitVcenter'] + bl['slitV']/2
        und = magFldCnt.arMagFld[0]
        srwlib.srwl.CalcStokesUR(stk, eBeam, und, arPrecPar)
        # intensity = stk.arS[0:npts]
        intensity = stk.to_int(radiation_polarisation)
    except:
         raise ValueError("Error running SRW.")

    return intensity, energy_array, time()-tzero


def tc_with_srwlibCalcElecFieldSR(bl: dict, 
                          eBeam: srwlib.SRWLPartBeam, 
                          magFldCnt: srwlib.SRWLMagFldC, 
                          energy_array: np.ndarray,
                          Kh: np.ndarray,
                          Kv: np.ndarray,
                          even_harmonics: bool,
                          h_slit_points: int, 
                          v_slit_points: int, 
                          radiation_characteristic: int, 
                          radiation_dependence: int, 
                          radiation_polarisation: int,
                          parallel: bool,
                          num_cores: int=None)-> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Calculates the intensity tuning curve using srwlib.CalcElecFieldSR.

    Args:
        bl (dict): Dictionary containing beamline parameters.
        eBeam (srwlib.SRWLPartBeam): Electron beam properties.
        magFldCnt (srwlib.SRWLMagFldC): Magnetic field container.
        energy_array (np.ndarray): Array of photon energies [eV].
        Kh (np.ndarray): Array of horizontal deflection parameters for each harmonic.
        Kv (np.ndarray): Array of vertical deflection parameters for each harmonic.
        even_harmonics (bool): Whether to include even harmonics in the calculation.
        h_slit_points (int): Number of horizontal slit points.
        v_slit_points (int): Number of vertical slit points.
        radiation_characteristic (int): Radiation characteristic:
               0 - "Single-Electron" Intensity;
               1 - "Multi-Electron" Intensity;
               4 - "Single-Electron" Radiation Phase;
               5 - Real part of Single-Electron Electric Field;
               6 - Imaginary part of Single-Electron Electric Field.
        radiation_dependence (int): Radiation dependence, e.g., 1 for angular distribution.
               0 - vs e (photon energy or time);
               1 - vs x (horizontal position or angle);
               2 - vs y (vertical position or angle);
               3 - vs x&y (horizontal and vertical positions or angles);
               4 - vs e&x (photon energy or time and horizontal position or angle);
               5 - vs e&y (photon energy or time and vertical position or angle);
               6 - vs e&x&y (photon energy or time, horizontal and vertical positions or angles).
        radiation_polarisation (int): Polarisation component to be extracted.
               0 - Linear Horizontal;
               1 - Linear Vertical;
               2 - Linear 45 degrees;
               3 - Linear 135 degrees;
               4 - Circular Right;
               5 - Circular Left;
               6 - Total.
        parallel (bool): Whether to use parallel computation.
        num_cores (int, optional): Number of CPU cores to use for parallel computation.
                                   Defaults to the number of available CPU cores.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: Total intensity (tuning curve), horizontal axis, and vertical axis.
    """

    nHarmMax = Kh.shape[1]

    if num_cores is None:
        num_cores = mp.cpu_count()

    if parallel:
        chunk_size = 20
        n_slices = len(energy_array)
        chunks = [(energy_array[i:i + chunk_size],
                   Kh[i:i + chunk_size, :],
                   Kv[i:i + chunk_size, :],
                   nHarmMax,
                   even_harmonics,
                   bl,
                   eBeam,
                   magFldCnt,
                   h_slit_points,
                   v_slit_points,
                   radiation_characteristic,
                   radiation_dependence,
                   radiation_polarisation,
                    ) for i in range(0, n_slices, chunk_size)]
        
        with mp.Pool() as pool:
            results = pool.map(core_tc_with_srwlibCalcElecFieldSR, chunks)
        tc = np.concatenate(results, axis=0)
    else:
        tc = core_tc_with_srwlibCalcElecFieldSR((energy_array,
                                                 Kh, 
                                                 Kv, 
                                                 nHarmMax,
                                                 even_harmonics,
                                                 bl,
                                                 eBeam,
                                                 magFldCnt,
                                                 h_slit_points,
                                                 v_slit_points,
                                                 radiation_characteristic,
                                                 radiation_dependence,
                                                 radiation_polarisation))
        
    if h_slit_points == 1 or v_slit_points == 1:
        h_axis = np.asarray([0])
        v_axis = np.asarray([0])
    else:
        h_axis = np.linspace(-bl['slitH']/2-bl['slitHcenter'], bl['slitH']/2-bl['slitHcenter'], h_slit_points)
        v_axis = np.linspace(-bl['slitV']/2-bl['slitVcenter'], bl['slitV']/2-bl['slitVcenter'], v_slit_points)
        
    return tc, h_axis, v_axis


def core_tc_with_srwlibCalcElecFieldSR(args):
    """
    Core function to calculate the intensity tuning curve using srwlib.CalcElecFieldSR.

    Args:
        args (tuple): Tuple containing:
            - energy_array (np.ndarray): Array of photon energies [eV].
            - Kh (np.ndarray): Horizontal deflection parameters for each harmonic.
            - Kv (np.ndarray): Vertical deflection parameters for each harmonic.
            - nHarmMax (int): Maximum number of harmonics.
            - even_harmonics (bool): Whether to include even harmonics.
            - bl (dict): Dictionary containing beamline parameters.
            - eBeam (srwlib.SRWPartBeam): Electron beam properties.
            - magFldCnt (srwlib.SRWLMagFldC): Magnetic field container.
            - h_slit_points (int): Number of horizontal slit points.
            - v_slit_points (int): Number of vertical slit points.
            - radiation_characteristic (int): Radiation characteristic.
            - radiation_dependence (int): Radiation dependence.
            - radiation_polarisation (int): Polarisation component.

    Returns:
        np.ndarray: Array containing the calculated intensity or flux.
    """

    energy, Kh, Kv, nHarmMax, even_harmonics, bl, eBeam, magFldCnt, \
        h_slit_points, v_slit_points, radiation_characteristic, radiation_dependence, \
            radiation_polarisation = args

    htc = np.zeros((len(energy), nHarmMax+1))

    for nharm in range(nHarmMax):
        if (nharm + 1) % 2 == 0 and even_harmonics or (nharm + 1) % 2 != 0:
            for i, dE in enumerate(energy):
                deflec_param = np.sqrt(Kh[i, nharm]**2 + Kv[i, nharm]**2)
                if deflec_param>0:
                    bl['Kv'] = Kv[i, nharm]
                    bl['Kh'] = Kh[i, nharm]                   
                    magFldCnt = set_magnetic_structure(bl, id_type='u')
                    htc[i, nharm+1], h_axis, v_axis = srwlibCalcElecFieldSR(
                                                bl, 
                                                eBeam, 
                                                magFldCnt,
                                                dE,
                                                h_slit_points=h_slit_points,
                                                v_slit_points=v_slit_points,
                                                radiation_characteristic=radiation_characteristic, 
                                                radiation_dependence=radiation_dependence,
                                                radiation_polarisation=radiation_polarisation,
                                                id_type='u',
                                                parallel=False,
                                                num_cores=1
                                                )
                    # htc[i, nharm+1] = (np.sum(flux)*(h_axis[1]-h_axis[0])*(v_axis[1]-v_axis[0]))*1E6   
    return htc


def tc_with_srwlibCalcStokesUR(bl: dict, 
                          eBeam: srwlib.SRWLPartBeam, 
                          magFldCnt: srwlib.SRWLMagFldC, 
                          energy_array: np.ndarray,
                          Kh: np.ndarray,
                          Kv: np.ndarray,
                          even_harmonics: bool,
                          h_slit_points: int, 
                          v_slit_points: int, 
                          radiation_characteristic: int, 
                          radiation_dependence: int, 
                          radiation_polarisation: int,
                          parallel: bool,
                          num_cores: int=None)-> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Calculates the intensity tuning curve using srwlib.CalcStokesUR.

    Args:
        bl (dict): Dictionary containing beamline parameters.
        eBeam (srwlib.SRWLPartBeam): Electron beam properties.
        magFldCnt (srwlib.SRWLMagFldC): Magnetic field container.
        energy_array (np.ndarray): Array of photon energies [eV].
        Kh (np.ndarray): Array of horizontal deflection parameters for each harmonic.
        Kv (np.ndarray): Array of vertical deflection parameters for each harmonic.
        even_harmonics (bool): Whether to include even harmonics in the calculation.
        h_slit_points (int): Number of horizontal slit points.
        v_slit_points (int): Number of vertical slit points.
        radiation_polarisation (int): Polarisation component to be extracted.
               0 - Linear Horizontal;
               1 - Linear Vertical;
               2 - Linear 45 degrees;
               3 - Linear 135 degrees;
               4 - Circular Right;
               5 - Circular Left;
               6 - Total.
        parallel (bool): Whether to use parallel computation.
        num_cores (int, optional): Number of CPU cores to use for parallel computation.
                                   Defaults to the number of available CPU cores.

    Returns:
        Tuple[np.ndarray]: Total intensity (tuning curve).
    """

    nHarmMax = Kh.shape[1]

    K = np.sqrt(Kh[:, 0]**2 + Kv[:, 0]**2)

    CalcStrokesEnergy = energy_array[K != 0]
    CalcStrokesEnergy /= CalcStrokesEnergy[0]*np.sqrt(2)
    CalcStrokesEnergy /= CalcStrokesEnergy[np.argmin(np.abs(CalcStrokesEnergy - 1))]

    if num_cores is None:
        num_cores = mp.cpu_count()

    if parallel:
        chunk_size = 20
        n_slices = len(energy_array)
        chunks = [(energy_array[i:i + chunk_size],
                   Kh[i:i + chunk_size, :],
                   Kv[i:i + chunk_size, :],
                   CalcStrokesEnergy,
                   nHarmMax,
                   even_harmonics,
                   bl,
                   eBeam,
                   magFldCnt,
                   h_slit_points,
                   v_slit_points,
                   radiation_characteristic,
                   radiation_dependence,
                   radiation_polarisation,
                    ) for i in range(0, n_slices, chunk_size)]
        
        with mp.Pool() as pool:
            results = pool.map(core_tc_with_srwlibCalcStokesUR, chunks)
        tc = np.concatenate(results, axis=0)
    else:
        tc = core_tc_with_srwlibCalcStokesUR((energy_array,
                                              Kh, 
                                              Kv, 
                                              CalcStrokesEnergy,
                                              nHarmMax,
                                              even_harmonics,
                                              bl,
                                              eBeam,
                                              magFldCnt,
                                              h_slit_points,
                                              v_slit_points,
                                              radiation_characteristic,
                                              radiation_dependence,
                                              radiation_polarisation))
               
    return tc

def core_tc_with_srwlibCalcStokesUR(args):
    """
    Core function to calculate the intensity tuning curve using srwlib.CalcStokesUR.

    Args:
        args (tuple): Tuple containing:
            - energy_array (np.ndarray): Array of photon energies [eV].
            - Kh (np.ndarray): Horizontal deflection parameters for each harmonic.
            - Kv (np.ndarray): Vertical deflection parameters for each harmonic.
            - nHarmMax (int): Maximum number of harmonics.
            - even_harmonics (bool): Whether to include even harmonics.
            - bl (dict): Dictionary containing beamline parameters.
            - eBeam (srwlib.SRWPartBeam): Electron beam properties.
            - magFldCnt (srwlib.SRWLMagFldC): Magnetic field container.
            - h_slit_points (int): Number of horizontal slit points.
            - v_slit_points (int): Number of vertical slit points.
            - radiation_characteristic (int): Radiation characteristic.
            - radiation_dependence (int): Radiation dependence.
            - radiation_polarisation (int): Polarisation component.

    Returns:
        np.ndarray: Array containing the calculated intensity or flux.
    """

    energy, Kh, Kv, CalcStrokesEnergy, nHarmMax, even_harmonics, bl, eBeam, magFldCnt, \
        h_slit_points, v_slit_points, radiation_characteristic, radiation_dependence, \
            radiation_polarisation = args

    htc = np.zeros((len(energy), nHarmMax+1))

    for nharm in range(nHarmMax):
        if (nharm + 1) % 2 == 0 and even_harmonics or (nharm + 1) % 2 != 0:
            for i, dE in enumerate(energy):
                deflec_param = np.sqrt(Kh[i, nharm]**2 + Kv[i, nharm]**2)
                if deflec_param>0:
                    bl['Kv'] = Kv[i, nharm]
                    bl['Kh'] = Kh[i, nharm]                   
                    magFldCnt = set_magnetic_structure(bl, id_type='u')
                    e_array = CalcStrokesEnergy*dE
                    flux = srwlibCalcStokesUR(
                                              bl, 
                                              eBeam, 
                                              magFldCnt,
                                              e_array,
                                              dE,
                                              h_slit_points=h_slit_points,
                                              v_slit_points=v_slit_points,
                                              radiation_polarisation=radiation_polarisation,
                                              parallel=False,
                                              num_cores=1
                                              )
                    # # RC 2024/12/04 - quick debug
                    # import matplotlib.pyplot as plt
                    # plt.plot(e_array, flux)
                    # plt.plot(dE, flux[np.where(e_array == dE)][0], "o")
                    # plt.show()
                    htc[i, nharm+1] = flux[np.where(e_array == dE)][0]  

    return htc


def tc_with_srwlibsrwl_wfr_emit_prop_multi_e(bl: dict, 
                          eBeam: srwlib.SRWLPartBeam, 
                          magFldCnt: srwlib.SRWLMagFldC, 
                          energy_array: np.ndarray,
                          Kh: np.ndarray,
                          Kv: np.ndarray,
                          even_harmonics: bool,
                          h_slit_points: int, 
                          v_slit_points: int, 
                          radiation_polarisation: int,
                          number_macro_electrons: int, 
                          aux_file_name: str,
                          parallel: bool,
                          num_cores: int=None)-> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Calculates the intensity tuning curve using srwlib.srwl_wfr_emit_prop_multi_e.

    Args:
        bl (dict): Dictionary containing beamline parameters.
        eBeam (srwlib.SRWLPartBeam): Electron beam properties.
        magFldCnt (srwlib.SRWLMagFldC): Magnetic field container.
        energy_array (np.ndarray): Array of photon energies [eV].
        Kh (np.ndarray): Array of horizontal deflection parameters for each harmonic.
        Kv (np.ndarray): Array of vertical deflection parameters for each harmonic.
        even_harmonics (bool): Whether to include even harmonics in the calculation.
        h_slit_points (int): Number of horizontal slit points.
        v_slit_points (int): Number of vertical slit points.
        radiation_polarisation (int): Polarisation component to be extracted.
               0 - Linear Horizontal;
               1 - Linear Vertical;
               2 - Linear 45 degrees;
               3 - Linear 135 degrees;
               4 - Circular Right;
               5 - Circular Left;
               6 - Total.
        number_macro_electrons (int): Total number of macro-electrons.
        aux_file_name (str): Auxiliary file name for saving intermediate data.
        parallel (bool): Whether to use parallel computation.
        num_cores (int, optional): Number of CPU cores to use for parallel computation.
                                   Defaults to the number of available CPU cores.

    Returns:
        Tuple[np.ndarray]: Total intensity (tuning curve), horizontal axis, and vertical axis.
    """

    nHarmMax = Kh.shape[1]

    if num_cores is None:
        num_cores = mp.cpu_count()

    if parallel:
        chunk_size = 20
        n_slices = len(energy_array)
        chunks = [(energy_array[i:i + chunk_size],
                   Kh[i:i + chunk_size, :],
                   Kv[i:i + chunk_size, :],
                   nHarmMax,
                   even_harmonics,
                   bl,
                   eBeam,
                   magFldCnt,
                   h_slit_points,
                   v_slit_points,
                   radiation_polarisation,
                   number_macro_electrons,
                   aux_file_name
                    ) for i in range(0, n_slices, chunk_size)]
        
        with mp.Pool() as pool:
            results = pool.map(core_tc_with_srwlibsrwl_wfr_emit_prop_multi_e, chunks)
        tc = np.concatenate(results, axis=0)
    else:
        tc = core_tc_with_srwlibsrwl_wfr_emit_prop_multi_e((energy_array,
                                                 Kh, 
                                                 Kv, 
                                                 nHarmMax,
                                                 even_harmonics,
                                                 bl,
                                                 eBeam,
                                                 magFldCnt,
                                                 h_slit_points,
                                                 v_slit_points,
                                                 radiation_polarisation,
                                                 number_macro_electrons,
                                                 aux_file_name))
        
    if h_slit_points == 1 or v_slit_points == 1:
        h_axis = np.asarray([0])
        v_axis = np.asarray([0])
    else:
        h_axis = np.linspace(-bl['slitH']/2-bl['slitHcenter'], bl['slitH']/2-bl['slitHcenter'], h_slit_points)
        v_axis = np.linspace(-bl['slitV']/2-bl['slitVcenter'], bl['slitV']/2-bl['slitVcenter'], v_slit_points)
        
    return tc, h_axis, v_axis


def core_tc_with_srwlibsrwl_wfr_emit_prop_multi_e(args):
    """
    Core function to calculate the intensity tuning curve using srwlib.srwl_wfr_emit_prop_multi_e.

    Args:
        args (tuple): Tuple containing:
            - energy_array (np.ndarray): Array of photon energies [eV].
            - Kh (np.ndarray): Horizontal deflection parameters for each harmonic.
            - Kv (np.ndarray): Vertical deflection parameters for each harmonic.
            - nHarmMax (int): Maximum number of harmonics.
            - even_harmonics (bool): Whether to include even harmonics.
            - bl (dict): Dictionary containing beamline parameters.
            - eBeam (srwlib.SRWPartBeam): Electron beam properties.
            - magFldCnt (srwlib.SRWLMagFldC): Magnetic field container.
            - h_slit_points (int): Number of horizontal slit points.
            - v_slit_points (int): Number of vertical slit points.
            - radiation_polarisation (int): Polarisation component.
            - number_macro_electrons (int): Total number of macro-electrons.
            - aux_file_name (str): Auxiliary file name for saving intermediate data.

    Returns:
        np.ndarray: Array containing the calculated intensity or flux.
    """
    energy, Kh, Kv, nHarmMax, even_harmonics, bl, eBeam, magFldCnt, \
        h_slit_points, v_slit_points, radiation_polarisation, number_macro_electrons, file_name = args
    
    htc = np.zeros((len(energy), nHarmMax+1))

    for nharm in range(nHarmMax):
        if (nharm + 1) % 2 == 0 and even_harmonics or (nharm + 1) % 2 != 0:
            for i, dE in enumerate(energy):
                deflec_param = np.sqrt(Kh[i, nharm]**2 + Kv[i, nharm]**2)
                if deflec_param>0:
                    bl['Kv'] = Kv[i, nharm]
                    bl['Kh'] = Kh[i, nharm]                   
                    magFldCnt = set_magnetic_structure(bl, id_type='u')
                     
                    htc[i, nharm+1], h_axis, v_axis = srwlibsrwl_wfr_emit_prop_multi_e(
                                                            bl, 
                                                            eBeam,
                                                            magFldCnt,
                                                            dE,
                                                            h_slit_points=h_slit_points,
                                                            v_slit_points=v_slit_points,
                                                            radiation_polarisation=radiation_polarisation,
                                                            id_type='u',
                                                            number_macro_electrons=number_macro_electrons,
                                                            aux_file_name=file_name,
                                                            parallel=False,
                                                            num_cores=1
                                                            )  

    return htc

#***********************************************************************************
# auxiliary functions accelerator functions
#***********************************************************************************

def get_undulator_max_harmonic_number(resonant_energy: float, photonEnergyMax: float) -> int:
    """
    Calculate the maximum harmonic number for an undulator to be considered by srwlib.CalcStokesUR.

    Args:
        resonant_energy (float): The resonance energy of the undulator [eV].
        photonEnergyMax (float): The maximum photon energy of interest [eV].

    Returns:
        int: The maximum harmonic number.
    """
    srw_max_harmonic_number = int(photonEnergyMax / resonant_energy * 2.5)
    if srw_max_harmonic_number < 15:
        srw_max_harmonic_number = 15
    return srw_max_harmonic_number





#!/bin/python

"""
This module provides a collection of functions for processing and analyzing data related
 to undulator radiation, power density, and spectra. It includes functions for reading 
 data from XOPPY files (HDF5 format), SPECTRA files (JSON format), and processing the 
 data to calculate various properties such as spectral power, cumulated power, 
 integrated power, and power density. Additionally, it offers functions for selecting 
 specific energy ranges within 3D data sets, spatially trimming data, and generating 
 animated GIFs of energy scans in the 3D data-sets. Overall, this module facilitates 
 the analysis and visualization of data obtained from simulations related to synchrotron 
 radiation sources.
"""

__author__ = ['Rafael Celestre']
__contact__ = 'rafael.celestre@synchrotron-soleil.fr'
__license__ = 'GPL-3.0'
__copyright__ = 'Synchrotron SOLEIL, Saint Aubin, France'
__created__ = '26/JAN/2024'
__changed__ = '25/NOV/2024'

import json
import multiprocessing
import os
import pickle
import sys
from typing import Any, Dict, List, Optional, Tuple, Union

import h5py as h5
import imageio
import matplotlib.pyplot as plt
import numpy as np
import scipy.integrate as integrate
from PIL import Image, ImageDraw, ImageFont
from scipy.constants import physical_constants
from scipy.interpolate import RegularGridInterpolator

try:
    import srwpy.srwlib as srwlib
    USE_SRWLIB = True
except:
    import oasys_srw.srwlib as srwlib
    USE_SRWLIB = True
if USE_SRWLIB is False:
     raise AttributeError("SRW is not available")

CHARGE = physical_constants["atomic unit of charge"][0]

#***********************************************************************************
# Spectrum
#***********************************************************************************

def write_spectrum(file_name: str, flux: np.array, energy: np.array) -> None:
    """
    Writes spectrum data to an HDF5 file.

    This function writes the provided energy and flux data to an HDF5 file. The data is stored 
    in the 'XOPPY_SPECTRUM' group within the file, with a subgroup for 'Spectrum'.

    Parameters:
        file_name (str): Base file path for saving the spectrum data. The file will be saved 
                         with the suffix '_spectrum.h5'.
        flux (np.array): 1D numpy array containing the flux data.
        energy (np.array): 1D numpy array containing the energy data.

    """
    if file_name is not None:
        with h5.File('%s_spectrum.h5'%file_name, 'w') as f:
            group = f.create_group('XOPPY_SPECTRUM')
            intensity_group = group.create_group('Spectrum')
            intensity_group.create_dataset('energy', data=energy)
            intensity_group.create_dataset('flux', data=flux) 

    spectral_power = flux*CHARGE*1E3

    cumulated_power = integrate.cumulative_trapezoid(spectral_power, energy, initial=0)
    integrated_power = integrate.trapezoid(spectral_power, energy)

    spectrumSRdict = {
        "spectrum":{
            "energy":energy,
            "flux": flux,
            "spectral_power": spectral_power,
            "cumulated_power": cumulated_power,
            "integrated_power": integrated_power
        }
    }

    return spectrumSRdict


def read_spectrum(file_list: List[str]) -> Dict:
    """
    Reads and processes spectrum data from files.

    This function reads spectrum data from files specified in 'file_list' and processes 
    it to compute spectral power, cumulated power, and integrated power.

    Parameters:
        file_list (List[str]): A list of file paths containing spectrum data.

    Returns:
        Dict: A dictionary containing processed spectrum data with the following keys:
            - 'spectrum': A dictionary containing various properties of the spectrum including:
                - 'energy': Array containing energy values.
                - 'flux': Array containing spectral flux data.
                - 'spectral_power': Array containing computed spectral power.
                - 'cumulated_power': Cumulated power computed using cumulative trapezoid integration.
                - 'integrated_power': Integrated power computed using trapezoid integration.
    """
    energy = []
    flux = []

    if isinstance(file_list, List) is False:
        file_list = [file_list]

    # new and official barc4sr format
    if file_list[0].endswith("h5") or file_list[0].endswith("hdf5"):
        for sim in file_list:
            print(sim)
            f = h5.File(sim, "r")
            energy = np.concatenate((energy, f["XOPPY_SPECTRUM"]["Spectrum"]["energy"][()]))
            flux = np.concatenate((flux, f["XOPPY_SPECTRUM"]["Spectrum"]["flux"][()]))

    # backwards compatibility - old barc4sr format
    elif file_list[0].endswith("pickle"):
        for sim in file_list:
            print(sim)
            f = open(sim, "rb")
            data = np.asarray(pickle.load(f))
            f.close()

            energy = np.concatenate((energy, data[0, :]))
            flux = np.concatenate((flux, data[1, :]))

    # SPECTRA format
    elif file_list[0].endswith("json"):
        for jsonfile in file_list:
            f = open(jsonfile)
            data = json.load(f)
            f.close()
            energy = np.concatenate((energy, data['Output']['data'][0]))
            flux = np.concatenate((flux, data['Output']['data'][1]))
            if 'Angular Flux Density' in data["Input"]["Configurations"]["Type"]:
                dist = data["Input"]["Configurations"]["Distance from the Source (m)"]
                flux /= (np.tan(0.5E-3)*dist*2*1E3)**2
    else:
        raise ValueError("Invalid file extension.")

    spectral_power = flux*CHARGE*1E3

    cumulated_power = integrate.cumulative_trapezoid(spectral_power, energy, initial=0)
    integrated_power = integrate.trapezoid(spectral_power, energy)

    spectrumSRdict = {
        "spectrum":{
            "energy":energy,
            "flux": flux,
            "spectral_power": spectral_power,
            "cumulated_power": cumulated_power,
            "integrated_power": integrated_power
        }
    }

    return spectrumSRdict

#***********************************************************************************
# Power density
#***********************************************************************************

def write_power_density(file_name: str, power_density: np.array, h_axis: np.array, v_axis: np.array) -> None:
    """
    Writes power density data to an HDF5 file.

    This function writes the provided power density data along with the corresponding 
    horizontal (h_axis) and vertical (v_axis) axes to an HDF5 file. The data is stored in 
    the 'XOPPY_POWERDENSITY' group within the file, with a subgroup for 'PowerDensity'.

    Parameters:
        file_name (str): Base file path for saving the power density data. The file will be 
                         saved with the suffix '_power_density.h5'.
        power_density (np.array): 2D numpy array containing the power density data.
        h_axis (np.array): 1D numpy array containing the horizontal axis data.
        v_axis (np.array): 1D numpy array containing the vertical axis data.

    """
    if file_name is not None:
        with h5.File('%s_power_density.h5' % file_name, 'w') as f:
            group = f.create_group('XOPPY_POWERDENSITY')
            sub_group = group.create_group('PowerDensity')
            sub_group.create_dataset('image_data', data=power_density)
            sub_group.create_dataset('axis_x', data=h_axis * 1e3)  # axis in [mm]
            sub_group.create_dataset('axis_y', data=v_axis * 1e3)

    dx = (h_axis[1]-h_axis[0])*1E3
    dy = (v_axis[1]-v_axis[0])*1E3

    CumPow = power_density.sum()*dx*dy

    print(f"Total received power: {CumPow:.3f} W")
    print(f"Peak power density: {power_density.max():.3f} W/mm^2")

    powDenSRdict = {
        "axis": {
            "x": h_axis,
            "y": v_axis,
            },
        "power_density": {
            "map":power_density,
            "CumPow": CumPow,
            "PowDenSRmax": power_density.max()
            }
        }
    
    return powDenSRdict

def read_power_density(file_name: str) -> Dict:
    """
    Reads power density data from an HDF5 (barc4sr) or JSON (SPECTRA) file and processes it.

    This function reads power density data from either an  HDF5 (barc4sr) or JSON (SPECTRA)
    file specified by 'file_name'. It extracts the power density map along with 
    corresponding x and y axes from the file.

    Parameters:
        file_name (str): File path containing power density data.

    Returns:
        Dict: A dictionary containing processed power density data with the following keys:
            - 'axis': A dictionary containing 'x' and 'y' axes arrays.
            - 'power_density': A dictionary containing power density-related data, including the power density map,
              total received power, and peak power density.
    """
    if file_name.endswith("h5") or file_name.endswith("hdf5"):
        f = h5.File(file_name, "r")
        PowDenSR = f["XOPPY_POWERDENSITY"]["PowerDensity"]["image_data"][()]

        x = f["XOPPY_POWERDENSITY"]["PowerDensity"]["axis_x"][()]
        y = f["XOPPY_POWERDENSITY"]["PowerDensity"]["axis_y"][()]

    elif file_name.endswith("json"):
        f = open(file_name)
        data = json.load(f)
        f.close()

        PowDenSR = np.reshape(data['Output']['data'][2],
                            (len(data['Output']['data'][1]), 
                             len(data['Output']['data'][0])))

        if "mrad" in data['Output']['units'][2]:
            dist = data["Input"]["Configurations"]["Distance from the Source (m)"]
            dx = (data["Input"]["Configurations"]["x Range (mm)"][1]-data["Input"]["Configurations"]["x Range (mm)"][0])*1e-3
            dy = (data["Input"]["Configurations"]["y Range (mm)"][1]-data["Input"]["Configurations"]["y Range (mm)"][0])*1e-3

            dtx = 2*np.arctan(dx/dist/2)*1e3    # mrad
            dty = 2*np.arctan(dy/dist/2)*1e3

            PowDenSR *= 1e3 * (dtx*dty)/(dx*dy*1e3*1e3)
            x = data['Output']['data'][0]
            y = data['Output']['data'][1]
        else:
            PowDenSR *= 1e3
    else:
        raise ValueError("Invalid file extension.")

    dx = x[1]-x[0]
    dy = y[1]-y[0]

    CumPow = PowDenSR.sum()*dx*dy

    print(f"Total received power: {CumPow:.3f} W")
    print(f"Peak power density: {PowDenSR.max():.3f} W/mm^2")

    powDenSRdict = {
        "axis": {
            "x": x,
            "y": y,
            },
        "power_density": {
            "map":PowDenSR,
            "CumPow": CumPow,
            "PowDenSRmax": PowDenSR.max()
            }
        }
    
    return powDenSRdict


def trim_and_resample_power_density(powDenSRdict: Dict, **kwargs: Union[float, bool]) -> Dict:
    """
    Trims and optionally resamples the power density data map.

    This function trims the power density data map based on specified criteria and optionally resamples
    it using interpolation if new axis values are provided. It returns the trimmed and resampled power density map
    along with cumulative power and maximum power density.

    Parameters:
        - powDenSRdict (Dict): A dictionary containing power density data with the following keys:
            - 'axis': A dictionary containing 'x' and 'y' axes arrays.
            - 'power_density': A dictionary containing the power density map.
        - **kwargs (Union[float, bool]): Additional keyword arguments for optional trimming and resampling:
            - 'dx' (float): Width in the x-direction for trimming.
            - 'dy' (float): Width in the y-direction for trimming.
            - 'xc' (float): Center of the trimming region along the x-axis.
            - 'yc' (float): Center of the trimming region along the y-axis.
            - 'X' (array_like): New x-axis values for resampling.
            - 'Y' (array_like): New y-axis values for resampling.

    Returns:
        Dict: A dictionary containing trimmed and resampled power density data with the following keys:
            - 'axis': A dictionary containing trimmed and resampled 'x' and 'y' axes arrays.
            - 'power_density': A dictionary containing the trimmed and resampled power density map,
              along with cumulative power and maximum power density.

    """

    PowDenSR = powDenSRdict["power_density"]["map"]

    x, y = powDenSRdict["axis"]["x"], powDenSRdict["axis"]["y"]
    xc, yc = 0, 0
    dx = x[-1] - x[0]
    dy = y[-1] - y[0]

    if kwargs:
        dx = kwargs.get("dx", dx)
        dy = kwargs.get("dy", dy)
        xc = kwargs.get("xc", xc)
        yc = kwargs.get("yc", yc)
        X = kwargs.get("X", x)
        Y = kwargs.get("Y", y)
        interpol = "X" in kwargs or "Y" in kwargs

    if interpol:
        print("Interpolation of PowDenSR")
        ygrid, xgrid = np.meshgrid(Y, X, indexing='ij')
        f = RegularGridInterpolator((y, x), PowDenSR, bounds_error=False, fill_value=0)
        PowDenSR = f((ygrid, xgrid))
        x, y = X, Y
    else:
        deltax, deltay = x[1] - x[0], y[1] - y[0]

        bool_x = np.logical_and((x - xc <=  dx / 2 + deltax / 20),
                                (x - xc >= -dx / 2 - deltax / 20))

        bool_y = np.logical_and((y - yc <=  dy / 2 + deltay / 20),
                                (y - yc >= -dy / 2 - deltay / 20)) 

        PowDenSR = PowDenSR[bool_y, :][:, bool_x]
        x = np.delete(x, np.logical_not(bool_x)) - xc
        y = np.delete(y, np.logical_not(bool_y)) - yc

    dx, dy = x[1] - x[0], y[1] - y[0]

    CumPow = PowDenSR.sum() * dx * dy

    print(f"Total received power: {CumPow:.3f} W")
    print(f"Peak power density: {PowDenSR.max():.3f} W/mm^2")

    return {
        "axis": {"x": x, "y": y},
        "power_density": {"map": PowDenSR, "CumPow": CumPow, "PowDenSRmax": PowDenSR.max()}
    }

#***********************************************************************************
# Spatial-spectral distribution
#***********************************************************************************

def write_emitted_radiation(file_name: str, intensity: np.array, energy: np.array, 
                            h_axis: np.array, v_axis: np.array, parallel_processing: bool = False) -> None:
    """
    Writes synchrotron radiation (I vs x vs y vs E) data to an HDF5 file.

    This function writes the provided intensity data along with the corresponding energy, 
    horizontal (h_axis), and vertical (v_axis) axes to an HDF5 file. The data is stored 
    in the 'XOPPY_RADIATION' group within the file, with a subgroup for 'Radiation'.

    Parameters:
        file_name (str): Base file path for saving the undulator radiation data. The file will be 
                         saved with the suffix '_undulator_radiation.h5'.
        intensity (np.array): 3D numpy array containing the intensity data.
        energy (np.array): 1D numpy array containing the energy axis data.
        h_axis (np.array): 1D numpy array containing the horizontal axis data.
        v_axis (np.array): 1D numpy array containing the vertical axis data.
        parallel_processing (bool, optional): Whether to use parallel processing. Defaults to False.

    """
    if file_name is not None:
        with h5.File('%s_undulator_radiation.h5' % file_name, 'w') as f:
            group = f.create_group('XOPPY_RADIATION')
            radiation_group = group.create_group('Radiation')
            radiation_group.create_dataset('stack_data', data=intensity)
            radiation_group.create_dataset('axis0', data=energy)
            radiation_group.create_dataset('axis1', data=h_axis * 1e3)
            radiation_group.create_dataset('axis2', data=v_axis * 1e3)

    if parallel_processing:
        return proc_spatial_spectral_dist_parallel(intensity, energy, h_axis, v_axis)
    else:
        return proc_spatial_spectral_dist(intensity, energy, h_axis, v_axis)


def read_emitted_radiation(file_list: List[str], parallel_processing: bool = False) -> Dict:
    """
    Reads synchrotron radiation emission (I vs x vs y vs E) data from from a list of files and processes it.

    This function reads emitted radiation data from a list of HDF5 files, concatenates the spectral 
    flux data, and processes it using either the proc_undulator_radiation function or the 
    proc_undulator_radiation_parallel function based on the value of parallel_processing.

    Parameters:
        - file_list (List[str]): A list of file paths containing synchrotron radiation data.
        - parallel_processing (bool, optional): Whether to use parallel processing. Defaults to False.

    Returns:
        Dict: A dictionary containing processed undulator radiation data.

    Notes:
        - The input HDF5 files should contain the following datasets:
            - 'XOPPY_RADIATION/Radiation/stack_data': 3D array representing the spectral flux data.
            - 'XOPPY_RADIATION/Radiation/axis0': 1D array representing the energy axis.
            - 'XOPPY_RADIATION/Radiation/axis1': 1D array representing the x-axis.
            - 'XOPPY_RADIATION/Radiation/axis2': 1D array representing the y-axis.
        - The spectral flux data from different files will be concatenated along the 0-axis.
        - The x and y axes are assumed to be the same for all files in the file_list.
    """

    if isinstance(file_list, List) is False:
        file_list = [file_list]
        
    energy = []
    spectral_flux_3D = []

    k = 0

    for sim in file_list:
        print(sim)
        f = h5.File(sim, "r")

        if k == 0:
            spectral_flux_3D = f["XOPPY_RADIATION"]["Radiation"]["stack_data"][()]
            k+=1
        else:
            spectral_flux_3D = np.concatenate((spectral_flux_3D, f["XOPPY_RADIATION"]["Radiation"]["stack_data"][()]), 0)
        energy = np.concatenate((energy, f["XOPPY_RADIATION"]["Radiation"]["axis0"][()]))

    print("UR files loaded")

    spectral_flux_3D = spectral_flux_3D.swapaxes(1, 2)

    x = f["XOPPY_RADIATION"]["Radiation"]["axis1"][()]
    y = f["XOPPY_RADIATION"]["Radiation"]["axis2"][()]

    if parallel_processing:
        return proc_spatial_spectral_dist_parallel(spectral_flux_3D, energy, x, y)
    else:
        return proc_spatial_spectral_dist(spectral_flux_3D, energy, x, y)


def proc_spatial_spectral_dist(spectral_flux_3D: np.ndarray, energy: np.ndarray, x: np.ndarray, y: np.ndarray) -> Dict:
    """
    Processes the synchotron radiation data.

    This function calculates various properties of synchotron radiation based on the provided 3D spectral flux data.

    Parameters:
        - spectral_flux_3D (np.ndarray): A 3D numpy array representing the spectral flux data. Shape: (n_slices, ny, nx)
        - energy (np.ndarray): A 1D numpy array containing the energy values.
        - x (np.ndarray): A 1D numpy array containing the x-axis values.
        - y (np.ndarray): A 1D numpy array containing the y-axis values.

    Returns:
        Dict: A dictionary containing processed synchotron radiation data with the following keys:
            - 'axis': Dictionary containing the x and y axis values.
            - 'spectral_power_3D': 3D numpy array of spectral power.
            - 'power_density': Dictionary containing the power density information with keys:
                - 'map': 2D numpy array representing the power density map.
                - 'CumPow': Cumulative power.
                - 'PowDenSRmax': Maximum power density.
            - 'spectrum': Dictionary containing spectral information with keys:
                - 'energy': 1D numpy array of energy values.
                - 'flux': Flux values.
                - 'spectral_power': Spectral power.
                - 'cumulated_power': Cumulative power.
                - 'integrated_power': Integrated power.

    Notes:
        - The input spectral_flux_3D should have dimensions (n_slices, ny, nx) where:
            - n_slices: Number of sample images.
            - ny: Number of points along the y-axis.
            - nx: Number of points along the x-axis.
        - The energy array should correspond to the energy values for each slice in spectral_flux_3D.
        - The x and y arrays represent the coordinates of the grid on which the data is sampled.

    """
    print("Processing undulator radiation")
    n_slices = spectral_flux_3D.shape[0]
    ny = spectral_flux_3D.shape[1]
    nx = spectral_flux_3D.shape[2]

    print(f"> {n_slices} sample images ({ny} x {nx}) found ({ sys.getsizeof(spectral_flux_3D) / (1024 ** 3):.2f} Gb in memory)")

    dx = (x[1]-x[0])
    dy = (y[1]-y[0])

    flux = dx*dy*np.sum(spectral_flux_3D, axis=(1, 2))
    spectral_power = flux*CHARGE*1E3
    cumulated_power = integrate.cumulative_trapezoid(spectral_power, energy, initial=0)
    integrated_power = integrate.trapezoid(spectral_power, energy)

    PowDenSR = integrate.trapezoid(spectral_flux_3D*CHARGE*1E3, energy, axis=0)

    CumPow = dx*dy*PowDenSR.sum()

    print(f"Puissance totale reçue : {CumPow:.3f} W")
    print(f"Puissance crête reçue (incidence normale): {PowDenSR.max():.3f} W/mm^2")

    URdict = {
        "axis": {
            "x": x,
            "y": y,
        },
        "spectral_power_3D":spectral_flux_3D*CHARGE*1E3,
        "power_density": {
            "map":PowDenSR,
            "CumPow": CumPow,
            "PowDenSRmax": PowDenSR.max()
        },
        "spectrum":{
            "energy":energy,
            "flux": flux,
            "spectral_power": spectral_power,
            "cumulated_power": cumulated_power,
            "integrated_power": integrated_power
        }
    }
    print("Dictionary written")
    return URdict


def proc_spatial_spectral_dist_parallel(spectral_flux_3D: np.ndarray, energy: np.ndarray, x: np.ndarray, y: np.ndarray, chunk_size: int = 25) -> Dict:
    """
    Process synchotron radiation data in parallel.

    This function calculates various properties of synchotron radiation based on the provided 3D spectral flux data.

    Parameters:
        - spectral_flux_3D (np.ndarray): A 3D numpy array representing the spectral flux data. Shape: (n_slices, ny, nx)
        - energy (np.ndarray): A 1D numpy array containing the energy values.
        - x (np.ndarray): A 1D numpy array containing the x-axis values.
        - y (np.ndarray): A 1D numpy array containing the y-axis values.
        - chunk_size (int, optional): Size of each chunk of spectral flux data for parallel processing. Default is 25.

    Returns:
        Dict: A dictionary containing processed synchotron radiation data with the following keys:
            - 'axis': Dictionary containing the x and y axis values.
            - 'spectral_power_3D': 3D numpy array of spectral power.
            - 'power_density': Dictionary containing the power density information with keys:
                - 'map': 2D numpy array representing the power density map.
                - 'CumPow': Cumulative power.
                - 'PowDenSRmax': Maximum power density.
            - 'spectrum': Dictionary containing spectral information with keys:
                - 'energy': 1D numpy array of energy values.
                - 'flux': Flux values.
                - 'spectral_power': Spectral power.
                - 'cumulated_power': Cumulative power.
                - 'integrated_power': Integrated power.

    Notes:
        - The input spectral_flux_3D should have dimensions (n_slices, ny, nx) where:
            - n_slices: Number of sample images.
            - ny: Number of points along the y-axis.
            - nx: Number of points along the x-axis.
        - The energy array should correspond to the energy values for each slice in spectral_flux_3D.
        - The x and y arrays represent the coordinates of the grid on which the data is sampled.

    """
    print("Processing undulator radiation (parallel)")
    n_slices = spectral_flux_3D.shape[0]
    ny = spectral_flux_3D.shape[1]
    nx = spectral_flux_3D.shape[2]

    print(f"> {n_slices} sample images ({ny} x {nx}) found ({ sys.getsizeof(spectral_flux_3D) / (1024 ** 3):.2f} Gb in memory)")

    dx = (x[1]-x[0])
    dy = (y[1]-y[0])

    # Divide the data into chunks
    chunks = [(spectral_flux_3D[i:i + chunk_size+1], energy[i:i + chunk_size+1], x, y) for i in range(0, n_slices, chunk_size)]
    
    # Create a multiprocessing Pool
    with multiprocessing.Pool() as pool:
        # Process each chunk in parallel
        processed_chunks = pool.map(process_chunk, chunks)
    
    # Concatenate the processed chunks
    PowDenSR = np.zeros((ny, nx))
    flux = []

    for i, (PowDenSR_chunk, flux_chunk, energy_chunck) in enumerate(processed_chunks):
        PowDenSR += PowDenSR_chunk
        if i == 0:
            flux.extend(flux_chunk)
            previous_energy_chunck = energy_chunck
        else: 
            if energy_chunck[0] == previous_energy_chunck[-1]:
                flux.extend(flux_chunk[1:])
            else:
                flux.extend(flux_chunk)
        previous_energy_chunck = energy_chunck

    spectral_power = np.asarray(flux)*CHARGE*1E3
    cumulated_power = integrate.cumulative_trapezoid(spectral_power, energy, initial=0)
    integrated_power = integrate.trapezoid(spectral_power, energy)
    CumPow = dx*dy*PowDenSR.sum()

    print(f"Puissance totale reçue : {CumPow:.3f} W")
    print(f"Puissance crête reçue (incidence normale): {PowDenSR.max():.3f} W/mm^2")

    URdict = {
        "axis": {
            "x": x,
            "y": y,
        },
        "spectral_power_3D":spectral_flux_3D*CHARGE*1E3,
        "power_density": {
            "map":PowDenSR,
            "CumPow": CumPow,
            "PowDenSRmax": PowDenSR.max()
        },
        "spectrum":{
            "energy":energy,
            "flux": np.asarray(flux),
            "spectral_power": spectral_power,
            "cumulated_power": cumulated_power,
            "integrated_power": integrated_power
        }
    }
    print("Dictionary written")
    return URdict


def process_chunk(args):
    """
    Process a chunk of spectral flux data.

    This function calculates the power density and flux for a chunk of spectral flux data.

    Parameters:
        args (tuple): A tuple containing the following elements:
            - spectral_flux_3D_chunk (np.ndarray): A chunk of spectral flux data.
            - energy_chunk (np.ndarray): A chunk of energy values.
            - x (np.ndarray): A 1D numpy array containing the x-axis values.
            - y (np.ndarray): A 1D numpy array containing the y-axis values.

    Returns:
        tuple: A tuple containing the calculated power density and flux.
    """
    spectral_flux_3D_chunk, energy_chunk, x, y = args

    dx = x[1] - x[0]
    dy = y[1] - y[0]

    flux = dx * dy * np.sum(spectral_flux_3D_chunk, axis=(1, 2))
    PowDenSR = integrate.trapezoid(spectral_flux_3D_chunk * CHARGE * 1E3, energy_chunk, axis=0)

    return PowDenSR, flux, energy_chunk


def select_energy_range(URdict: Dict, ei: float, ef: float, **kwargs: Union[float, bool]) -> Dict:
    """
    Selects a specific energy range from the synchotron radiation data and returns processed data within that range.

    This function selects a specific energy range from the given synchotron radiation data dictionary (URdict)
    and returns processed data within that range. Optionally, it allows trimming the data based on specified criteria.

    Parameters:
        - URdict (Dict): A dictionary containing synchotron radiation data with the following keys:
            - 'axis': A dictionary containing 'x' and 'y' axes arrays.
            - 'spectrum': A dictionary containing energy-related data, including 'energy' array.
            - 'spectral_power_3D': A 3D array representing spectral power density.
        - ei (float): The initial energy of the selected range.
        - ef (float): The final energy of the selected range.
        - **kwargs (Union[float, bool]): Additional keyword arguments for optional trimming:
            - 'dx' (float): Width in the x-direction for trimming.
            - 'dy' (float): Width in the y-direction for trimming.
            - 'xc' (float): Center of the trimming region along the x-axis.
            - 'yc' (float): Center of the trimming region along the y-axis.

    Returns:
        Dict: A dictionary containing processed synchotron radiation data within the selected energy range,
        trimmed based on the specified criteria (if any).

    Notes:
        - If 'ei' or 'ef' is set to -1, the function selects the minimum or maximum energy from the available data, respectively.
        - If 'ei' is equal to 'ef', the function duplicates the data for that energy and increments the energy values by 1.

    """
    x = URdict["axis"]["x"]
    y = URdict["axis"]["y"]

    dx = kwargs.get("dx", x[-1] - x[0])
    dy = kwargs.get("dy", y[-1] - y[0])
    xc = kwargs.get("xc", 0)
    yc = kwargs.get("yc", 0)

    deltax, deltay = x[1] - x[0], y[1] - y[0]

    bool_x = np.logical_and((x - xc <= dx / 2 + deltax / 20),
                            (x - xc >= -dx / 2 - deltax / 20))

    bool_y = np.logical_and((y - yc <= dy / 2 + deltay / 20),
                            (y - yc >= -dy / 2 - deltay / 20))

    if ei == -1:
        ei = URdict["spectrum"]["energy"][0]
    if ef == -1:
        ef = URdict["spectrum"]["energy"][-1]

    crop_map = np.logical_not(np.logical_and((URdict["spectrum"]["energy"] <= ef),
                                             (URdict["spectrum"]["energy"] >= ei)))
    energy = np.delete(URdict["spectrum"]["energy"], crop_map)
    spectral_flux_3D = np.delete(URdict["spectral_power_3D"], crop_map, axis=0) / (CHARGE * 1E3)

    if ei == ef:
        spectral_flux_3D = np.concatenate((spectral_flux_3D, spectral_flux_3D), axis=0)
        energy = np.concatenate((energy, energy + 1))

    if kwargs:  # Check if trimming is requested
        spectral_flux_3D = spectral_flux_3D[:, bool_y, :][:, :, bool_x]
        x = np.delete(x, np.logical_not(bool_x)) - xc
        y = np.delete(y, np.logical_not(bool_y)) - yc

    return proc_spatial_spectral_dist_parallel(spectral_flux_3D, energy, x, y)


def animate_energy_scan(URdict: Dict, file_name: str, **kwargs: Any) -> None:
    """
    Generate an animated GIF of an energy scan.

    Args:
        URdict (Dict): Dictionary containing spectral power data.
        file_name (str): Name of the output GIF file.
        **kwargs: Additional keyword arguments.
            duration_per_frame (float): Duration of each frame in seconds. Defaults to 0.05.
            frame_rate (float): Frame rate in frames per second. Overrides duration_per_frame if provided.
            cmap (str): Colormap for visualization. Defaults to 'plasma'.
            cumSum (bool): If True, generate an additional GIF with cumulative sum. Defaults to False.
            ScaleBar (bool): If True, add a scale bar to the GIF. Defaults to False.
            ScaleBarLength (float): Length of the scale bar in the unit of theScaleBarUnit.
            ScaleBarUnit (str): Unit of measurement for the scale bar.
            group (bool): If True, group the energy scan with the cumulative sum. Defaults to False.

    """
    duration_per_frame = 0.05
    cmap = 'plasma'
    cumSum = False
    ScaleBar = False
    group = False

    if bool(kwargs):

        if "duration_per_frame" in kwargs.keys():
            duration_per_frame = kwargs["duration_per_frame"]
        if "frame_rate" in kwargs.keys():
            duration_per_frame = 1/kwargs["frame_rate"]
        if "cmap" in kwargs.keys():
            cmap = kwargs["cmap"]
        if "cumSum" in kwargs.keys():
            cumSum = kwargs["cumSum"]

        ScaleBarLength = None
        ScaleBarUnit = None
        if "ScaleBar" in kwargs.keys():
            ScaleBar = kwargs["ScaleBar"]
        if "ScaleBarLength" in kwargs.keys():
            ScaleBar = True
            ScaleBarLength = kwargs["ScaleBarLength"]
        if "ScaleBarUnit" in kwargs.keys():
            ScaleBar = True
            ScaleBarUnit = kwargs["ScaleBarUnit"]

        if ScaleBar is True and ScaleBarUnit is None:
            # warnings.warn(">> Scale bar unit nor provided. No scale bar will be displayed", Warning)
            print(">> Warning: Scale bar unit nor provided. No scale bar will be displayed")
            ScaleBar = False

        if ScaleBar:
            if ScaleBarLength is None:
                dh = np.round((URdict["axis"]["x"][-1]-URdict["axis"]["x"][0])/4)
                ScaleBarLength = dh-dh%2
            PixelsPerLengthUnit = len(URdict["axis"]["x"])/(URdict["axis"]["x"][-1]-URdict["axis"]["x"][0])
            ScaleLengthPixels = int(ScaleBarLength * PixelsPerLengthUnit)
        
        if "group" in kwargs.keys():
            group = kwargs["group"]

    global_min = np.min(URdict["spectral_power_3D"])
    global_max = np.max(URdict["spectral_power_3D"])

    if group:
        cumulated_power = np.cumsum(URdict["spectral_power_3D"], axis=0)
        with imageio.get_writer(file_name + ".gif", mode='I', duration=duration_per_frame) as writer:
            for i, frame in enumerate(URdict["spectral_power_3D"]):

                cp_frame_min = np.min(cumulated_power[i, :, :])
                cp_frame_max = np.max(cumulated_power[i, :, :])
                cp_frame_normalized = (cumulated_power[i, :, :] - cp_frame_min) / (cp_frame_max - cp_frame_min)

                dnan = 5
                frame_normalized = (frame - global_min) / (global_max - global_min)
                frame_normalized = np.hstack((frame_normalized, np.full((frame_normalized.shape[0], dnan), np.nan)))
                frame_normalized = np.concatenate((frame_normalized, cp_frame_normalized), axis=1)

                frame_colored = plt.cm.get_cmap(cmap)(frame_normalized)
                # Convert the colormap to uint8 format (0-255)
                frame_colored_uint8 = (frame_colored[:, :, :3] * 255).astype(np.uint8)
                
                # Convert NumPy array to PIL Image
                pil_image = Image.fromarray(frame_colored_uint8)
                
                # Add frame number as text overlay
                draw = ImageDraw.Draw(pil_image)
                font = ImageFont.truetype("arial.ttf", 20)
                draw.text((15, 15), f"E = {URdict['spectrum']['energy'][i]:.2f} eV", fill=(255, 255, 255), font=font)
                draw.text((15+frame.shape[1]+dnan, 15), "cummulated", fill=(255, 255, 255), font=font)

                if ScaleBar:
                    # Add scale bar
                    scale_bar_x0 = pil_image.width - 10 - ScaleLengthPixels
                    scale_bar_y0 = pil_image.height - 20
                    scale_bar_x1 = pil_image.width - 10
                    scale_bar_y1 = pil_image.height - 15
                    draw.rectangle([scale_bar_x0, scale_bar_y0, scale_bar_x1, scale_bar_y1], fill=(255, 255, 255))
                    # Add text for scale bar length
                    scale_text = f"{ScaleBarLength} {ScaleBarUnit}"
                    text_left, text_top, text_right, text_bottom = draw.textbbox(xy=(0,0), text=scale_text, font=font)
                    text_width, text_height = (text_right - text_left, text_bottom - text_top)
                    text_x = scale_bar_x0 + (scale_bar_x1 - scale_bar_x0 - text_width) // 2
                    text_y = scale_bar_y0 - text_height - 10
                    draw.text((text_x, text_y), scale_text, fill=(255, 255, 255), font=font)

                # Convert PIL Image back to NumPy array
                frame_with_tag = np.array(pil_image)
                
                writer.append_data(frame_with_tag)

        print(f"GIF created successfully: {file_name.split('/')[-1]}")
    else:
        with imageio.get_writer(file_name + ".gif", mode='I', duration=duration_per_frame) as writer:
            for i, frame in enumerate(URdict["spectral_power_3D"]):

                frame_normalized = (frame - global_min) / (global_max - global_min)
                frame_colored = plt.cm.get_cmap(cmap)(frame_normalized)

                # Convert the colormap to uint8 format (0-255)
                frame_colored_uint8 = (frame_colored[:, :, :3] * 255).astype(np.uint8)
                
                # Convert NumPy array to PIL Image
                pil_image = Image.fromarray(frame_colored_uint8)
                
                # Add frame number as text overlay
                draw = ImageDraw.Draw(pil_image)
                font = ImageFont.truetype("arial.ttf", 20)
                draw.text((15, 15), f"E = {URdict['spectrum']['energy'][i]:.2f} eV", fill=(255, 255, 255), font=font)
                        
                if ScaleBar:
                    # Add scale bar
                    scale_bar_x0 = pil_image.width - 10 - ScaleLengthPixels
                    scale_bar_y0 = pil_image.height - 20
                    scale_bar_x1 = pil_image.width - 10
                    scale_bar_y1 = pil_image.height - 15
                    draw.rectangle([scale_bar_x0, scale_bar_y0, scale_bar_x1, scale_bar_y1], fill=(255, 255, 255))
                    # Add text for scale bar length
                    scale_text = f"{ScaleBarLength} {ScaleBarUnit}"
                    text_left, text_top, text_right, text_bottom = draw.textbbox(xy=(0,0), text=scale_text, font=font)
                    text_width, text_height = (text_right - text_left, text_bottom - text_top)
                    text_x = scale_bar_x0 + (scale_bar_x1 - scale_bar_x0 - text_width) // 2
                    text_y = scale_bar_y0 - text_height - 10
                    draw.text((text_x, text_y), scale_text, fill=(255, 255, 255), font=font)

                # Convert PIL Image back to NumPy array
                frame_with_tag = np.array(pil_image)
                
                writer.append_data(frame_with_tag)

        print(f"GIF created successfully: {file_name.split('/')[-1]}")

        if cumSum:
            cumulated_power = np.cumsum(URdict["spectral_power_3D"], axis=0)
            with imageio.get_writer(file_name + "_CumSum.gif", mode='I', duration=duration_per_frame) as writer:
                for i, frame in enumerate(cumulated_power):

                    frame_min = np.min(frame)
                    frame_max = np.max(frame)

                    frame_normalized = (frame - frame_min) / (frame_max - frame_min)
                    frame_colored = plt.cm.get_cmap(cmap)(frame_normalized)

                    # Convert the colormap to uint8 format (0-255)
                    frame_colored_uint8 = (frame_colored[:, :, :3] * 255).astype(np.uint8)
                    
                    # Convert NumPy array to PIL Image
                    pil_image = Image.fromarray(frame_colored_uint8)
                    
                    # Add frame number as text overlay
                    draw = ImageDraw.Draw(pil_image)
                    font = ImageFont.truetype("arial.ttf", 20)
                    draw.text((15, 15), f"E = {URdict['spectrum']['energy'][i]:.2f} eV (cummulated)", fill=(255, 255, 255), font=font)
                    if ScaleBar:
                        # Add scale bar
                        scale_bar_x0 = pil_image.width - 10 - ScaleLengthPixels
                        scale_bar_y0 = pil_image.height - 20
                        scale_bar_x1 = pil_image.width - 10
                        scale_bar_y1 = pil_image.height - 15
                        draw.rectangle([scale_bar_x0, scale_bar_y0, scale_bar_x1, scale_bar_y1], fill=(255, 255, 255))
                        # Add text for scale bar length
                        scale_text = f"{ScaleBarLength} {ScaleBarUnit}"
                        text_left, text_top, text_right, text_bottom = draw.textbbox(xy=(0,0), text=scale_text, font=font)
                        text_width, text_height = (text_right - text_left, text_bottom - text_top)
                        text_x = scale_bar_x0 + (scale_bar_x1 - scale_bar_x0 - text_width) // 2
                        text_y = scale_bar_y0 - text_height - 10
                        draw.text((text_x, text_y), scale_text, fill=(255, 255, 255), font=font)
                    # Convert PIL Image back to NumPy array
                    frame_with_tag = np.array(pil_image)
                    
                    writer.append_data(frame_with_tag)

            print(f"GIF created successfully: {file_name.split('/')[-1]+'_CumSum'}")

#***********************************************************************************
# Tuning curves
#***********************************************************************************

def write_tuning_curve(file_name: str, flux: np.array, Kh: np.array, Kv: np.array, energy: np.array) -> None:
    """
    Writes tuning curve data to an HDF5 file.

    This function writes the provided energy and flux data to an HDF5 file. The data is stored 
    in the 'XOPPY_SPECTRUM' group within the file, with a subgroup for 'TC'.

    Parameters:
        file_name (str): Base file path for saving the tuning curve data. The file will be saved 
                         with the suffix '_tc.h5'.
        flux (np.array): 1D numpy array containing the flux data.
        energy (np.array): 1D numpy array containing the energy data.

    """
    if file_name is not None:
        with h5.File('%s_tc.h5'%file_name, 'w') as f:
            group = f.create_group('XOPPY_SPECTRUM')
            intensity_group = group.create_group('TC')
            intensity_group.create_dataset('energy', data=energy)
            intensity_group.create_dataset('flux', data=flux) 
            intensity_group.create_dataset('Kh', data=Kh) 
            intensity_group.create_dataset('Kv', data=Kv) 

    tcSRdict = {
            "energy": energy,
            "flux": flux,
            "Kh": Kh,
            "Kv": Kv
    }

    return tcSRdict

def read_tuning_curve(file_name: str) -> Dict:
    """
    Reads and processes tuning curve data from files.

    Parameters:
        file_name (str): A file path containing tuning curve data.

    Returns:
        Dict: A dictionary containing processed tuning curve data with the following keys:
            - 'TC': A dictionary containing various properties of the tuning curve including:
                - 'energy': Array containing energy values.
                - 'flux': Array containing spectral flux data.
    """

    if file_name.endswith("h5") or file_name.endswith("hdf5"):
        print(file_name)
        with h5.File(file_name, "r") as f:
            energy =  f["XOPPY_SPECTRUM"]["TC"]["energy"][()]
            flux = f["XOPPY_SPECTRUM"]["TC"]["flux"][()]
            Kh = f["XOPPY_SPECTRUM"]["TC"]["Kh"][()]
            Kv = f["XOPPY_SPECTRUM"]["TC"]["Kv"][()]

    tcSRdict = {
            "energy": energy,
            "flux": flux,
            "Kh": Kh,
            "Kv": Kv
    }

    return tcSRdict

#***********************************************************************************
# Wavevfront
#***********************************************************************************
   
def write_wavefront(file_name: str, intensity:np.array, phase:np.array, h_axis:np.array,
                    v_axis:np.array) -> None:
    """
    Writes wavefront data to an HDF5 file.

    This function writes the provided intensity and phase maps along with the corresponding 
    horizontal (h_axis) and vertical (v_axis) axes to an HDF5 file. The data is stored in the 
    'XOPPY_WAVEFRONT' group within the file, with subgroups for 'Intensity' and 'Phase'.

    Parameters:
        file_name (str): Base file path for saving the wavefront data. The file will be saved with 
                         the suffix '_undulator_wft.h5'.
        intensity (np.array): 2D numpy array containing the intensity data.
        phase (np.array): 2D numpy array containing the phase data.
        h_axis (np.array): 1D numpy array containing the horizontal axis data.
        v_axis (np.array): 1D numpy array containing the vertical axis data.

    """
    if file_name is not None:
        with h5.File('%s_undulator_wft.h5'%file_name, 'w') as f:
            group = f.create_group('XOPPY_WAVEFRONT')
            intensity_group = group.create_group('Intensity')
            intensity_group.create_dataset('image_data', data=intensity)
            intensity_group.create_dataset('axis_x', data=h_axis*1e3) 
            intensity_group.create_dataset('axis_y', data=v_axis*1e3)
            intensity_group = group.create_group('Phase')
            intensity_group.create_dataset('image_data', data=phase)
            intensity_group.create_dataset('axis_x', data=h_axis*1e3) 
            intensity_group.create_dataset('axis_y', data=v_axis*1e3)

    wftDict = {
        "axis": {
            "x": h_axis,
            "y": v_axis,
            },
        "wavefront": {
            "intensity":intensity,
            "phase": phase,
            }
        }
    
    return wftDict

def read_wavefront(file_name: str) -> Dict:
    """
    Reads wavefront data from an HDF5  file and processes it.

    This function reads wavefront data from an HDF5 file (barc4sr) file specified by 
    'file_name'. It extracts the intensity and phase maps along with  corresponding x 
    and y axes from the file.

    Parameters:
        file_name (str): File path containing wavefront data.

    Returns:
        Dict: A dictionary containing processed wavefront data with the following keys:
            - 'axis': A dictionary containing 'x' and 'y' axes arrays.
            - 'intensity': 2D numpy array of intensity data.
            - 'phase': 2D numpy array of phase data.
    """
    if file_name.endswith("h5") or file_name.endswith("hdf5"):
        f = h5.File(file_name, "r")
        phase = f["XOPPY_WAVEFRONT"]["Phase"]["image_data"][()]
        intensity = f["XOPPY_WAVEFRONT"]["Intensity"]["image_data"][()]

        x = f["XOPPY_WAVEFRONT"]["Phase"]["axis_x"][()]
        y = f["XOPPY_WAVEFRONT"]["Phase"]["axis_y"][()]
        
    wftDict = {
        "axis": {
            "x": x,
            "y": y,
            },
        "wavefront": {
            "intensity":intensity,
            "phase": phase,
            }
        }
    
    return wftDict

#***********************************************************************************
# electron trajectory
#***********************************************************************************

def write_electron_trajectory(file_name:str, eTraj: srwlib.SRWLPrtTrj):
    """
    Saves electron trajectory data to an HDF5 file and returns a dictionary containing the trajectory data.

    This function processes the trajectory data from an `SRWLPrtTrj` object and stores it in both an HDF5 file 
    and a Python dictionary. 

    Parameters:
        file_name (str): Base file path for saving the trajectory data. The data will be saved 
                         in a file with the suffix '_eTraj.h5'.
        eTraj (SRWLPrtTrj): SRW library object containing the electron trajectory data. The object must include:
            - `arX`: Array of horizontal positions [m].
            - `arXp`: Array of horizontal relative velocities (trajectory angles) [rad].
            - `arY`: Array of vertical positions [m].
            - `arYp`: Array of vertical relative velocities (trajectory angles) [rad].
            - `arZ`: Array of longitudinal positions [m].
            - `arZp`: Array of longitudinal relative velocities (trajectory angles) [rad].
            - `arBx` (optional): Array of horizontal magnetic field components [T].
            - `arBy` (optional): Array of vertical magnetic field components [T].
            - `arBz` (optional): Array of longitudinal magnetic field components [T].
            - `np`: Number of trajectory points.
            - `ctStart`: Start value of the independent variable (c*t) for the trajectory [m].
            - `ctEnd`: End value of the independent variable (c*t) for the trajectory [m].

    Returns:
        dict: A dictionary containing the trajectory data with the following keys:
              - "ct": List of time values corresponding to the trajectory points.
              - "X", "Y", "Z": Lists of positions in the respective axes.
              - "BetaX", "BetaY", "BetaZ": Lists of velocity components (trajectory angles) in the respective axes.
              - "Bx", "By", "Bz" (optional): Lists of magnetic field components in the respective axes, if present.
    """

    eTrajDict = {"eTraj":{
        "ct": [],
        "X": [],
        "BetaX": [],
        "Y": [],
        "BetaY": [],
        "Z": [],
        "BetaZ": [],
    }}

    if hasattr(eTraj, 'arBx'):
        eTrajDict["eTraj"]["Bx"] = []
    if hasattr(eTraj, 'arBy'):
        eTrajDict["eTraj"]["By"] = []
    if hasattr(eTraj, 'arBz'):
        eTrajDict["eTraj"]["Bz"] = []

    if file_name is not None:
        with h5.File(f"{file_name}_eTraj.h5", "w") as f:
            group = f.create_group("XOPPY_ETRAJ")
            intensity_group = group.create_group("eTraj")
            
            intensity_group.create_dataset("ct", data=np.zeros(eTraj.np))
            intensity_group.create_dataset("X", data=eTraj.arX)
            intensity_group.create_dataset("BetaX", data=eTraj.arXp)
            intensity_group.create_dataset("Y", data=eTraj.arY)
            intensity_group.create_dataset("BetaY", data=eTraj.arYp)
            intensity_group.create_dataset("Z", data=eTraj.arZ)
            intensity_group.create_dataset("BetaZ", data=eTraj.arZp)
            if hasattr(eTraj, 'arBx'):
                intensity_group.create_dataset("Bx", data=eTraj.arBx)
            if hasattr(eTraj, 'arBy'):
                intensity_group.create_dataset("By", data=eTraj.arBy)
            if hasattr(eTraj, 'arBz'):
                intensity_group.create_dataset("Bz", data=eTraj.arBz)

    eTrajDict["eTraj"]["ct"] = np.zeros(eTraj.np)
    eTrajDict["eTraj"]["X"] = np.asarray(eTraj.arX)
    eTrajDict["eTraj"]["BetaX"] = np.asarray(eTraj.arXp)
    eTrajDict["eTraj"]["Y"] = np.asarray(eTraj.arY)
    eTrajDict["eTraj"]["BetaY"] = np.asarray(eTraj.arY)
    eTrajDict["eTraj"]["Z"] = np.asarray(eTraj.arZ)
    eTrajDict["eTraj"]["BetaZ"] = np.asarray(eTraj.arZ)
    eTrajDict["eTraj"]["Bx"] = np.asarray(eTraj.arBx)
    eTrajDict["eTraj"]["By"] = np.asarray(eTraj.arBy)
    eTrajDict["eTraj"]["Bz"] = np.asarray(eTraj.arBz)

    return eTrajDict
    

def read_electron_trajectory(file_path: str) -> Dict[str, List[Union[float, None]]]:
    """
    Reads SRW electron trajectory data from a .h5 file (XOPPY_ETRAJ format).

    Args:
        file_path (str): The path to the .h5 file containing electron trajectory data.

    Returns:
        dict: A dictionary where keys are the column names (ct, X, BetaX, Y, BetaY, Z, BetaZ, Bx, By, Bz),
            and values are lists containing the corresponding column data from the file.
    """
    result = {"eTraj": {}}

    with h5.File(file_path, "r") as f:
        try:
            trajectory_group = f["XOPPY_ETRAJ"]["eTraj"]
        except KeyError:
            raise ValueError(f"Invalid file structure: {file_path} does not contain 'XOPPY_ETRAJ/eTraj'.")

        # Read datasets
        for key in trajectory_group.keys():
            result["eTraj"][key] = trajectory_group[key][:].tolist()

    return result


def read_electron_trajectory_dat(file_path: str) -> Dict[str, List[Union[float, None]]]:
    """
    Reads SRW electron trajectory data from a .dat file (SRW native format).

    Args:
        file_path (str): The path to the .dat file containing electron trajectory data.

    Returns:
        dict: A dictionary where keys are the column names extracted from the header
            (ct, X, BetaX, Y, BetaY, Z, BetaZ, Bx, By, Bz),
            and values are lists containing the corresponding column data from the file.
    """
    data = []
    header = None
    with open(file_path, 'r') as file:
        header_line = next(file).strip()
        header = [col.split()[0] for col in header_line.split(',')]
        header[0] = header[0].replace("#","")
        for line in file:
            values = line.strip().split('\t')
            values = [float(value) if value != '' else None for value in values]
            data.append(values)
            
    eTrajDict = {}
    for i, key in enumerate(header):
        eTrajDict[key] = np.asarray([row[i] for row in data])

    return eTrajDict

#***********************************************************************************
# magnetic measurments
#***********************************************************************************

def write_magnetic_field(mag_field_array: np.ndarray, file_path: Optional[str] = None) -> srwlib.SRWLMagFld3D:
    """
    Generate a 3D magnetic field object based on the input magnetic field array.

    Parameters:
        mag_field_array (np.ndarray): Array containing magnetic field data. Each row corresponds to a point in the 3D space,
                                      where the first column represents the position along the longitudinal axis, and subsequent 
                                      columns represent magnetic field components (e.g., Bx, By, Bz).
        file_path (str, optional): File path to save the generated magnetic field object. If None, the object won't be saved.

    Returns:
        SRWLMagFld3D: Generated 3D magnetic field object.

    """
    nfield, ncomponents = mag_field_array.shape

    field_axis = (mag_field_array[:, 0] - np.mean(mag_field_array[:, 0])) * 1e-3

    Bx = mag_field_array[:, 1]
    if ncomponents > 2:
        By = mag_field_array[:, 2]
    else:
        By = np.zeros(nfield)
    if ncomponents > 3:
        Bz = mag_field_array[:, 3]
    else:
        Bz = np.zeros(nfield)

    magFldCnt = srwlib.SRWLMagFld3D(Bx, By, Bz, 1, 1, nfield - 1, 0, 0, field_axis[-1]-field_axis[0], 1)

    if file_path is not None:
        print(f">>> saving {file_path}")
        magFldCnt.save_ascii(file_path)

    return magFldCnt


def read_magnetic_measurement(file_path: str) -> np.ndarray:
    """
    Read magnetic measurement data from a file.

    Parameters:
        file_path (str): The path to the file containing magnetic measurement data.

    Returns:
        np.ndarray: A NumPy array containing the magnetic measurement data.
    """

    data = []

    with open(file_path, 'r') as file:
        for line in file:
            if not line.startswith('#'):
                values = line.split( )
                data.append([float(value) for value in values])
                
    return np.asarray(data)


#***********************************************************************************
# Mutual intensity, (Cross-) Spectral Density and Degree of Coherence
#***********************************************************************************

if __name__ == '__main__':

    file_name = os.path.basename(__file__)

    print(f"This is the barc4sr.{file_name} module!")
    print("This module provides functions for processing and analyzing data related to synchrotron radiation, power density, and spectra.")

